from typing import List, NamedTuple

import structlog

from spark_pipeline_framework.utilities.helix_geolocation.v1.raw_address import (
    RawAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.standardized_address import (
    StandardizedAddress,
)

logger = structlog.get_logger(__file__)


class CacheResult(NamedTuple):
    found: List[StandardizedAddress]
    not_found: List[RawAddress]



from pathlib import Path

from pyspark.sql import SparkSession, DataFrame

from spark_pipeline_framework.transformers.fhir_text_reader.v1.fhir_text_reader import (
    FhirTextReader,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)


def test_fhir_text_reader(spark_session: SparkSession) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath("./")

    test_files_dir: Path = data_dir.joinpath("test_files/patients")

    df: DataFrame = create_empty_dataframe(spark_session=spark_session)

    # Act
    FhirTextReader(
        view="patients", file_path=test_files_dir, destination_column_name="bundle"
    ).transform(df)

    # Assert
    result_df: DataFrame = spark_session.table("patients")
    result_df.printSchema()
    result_df.select("bundle").show(truncate=False)

    assert result_df.count() == 2

    assert (
        result_df.select("bundle")
        .collect()[0][0]
        .startswith('{"resourceType":"Patient"')
    )



from typing import Optional, Dict, Any

from spark_pipeline_framework.logger.log_level import LogLevel
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.pretty_print import (
    get_pretty_data_frame,
    get_data_frame_as_csv,
)


class FrameworkSqlTransformer(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        sql: Optional[str] = None,
        view: Optional[str] = None,
        log_sql: bool = False,
        log_result: bool = False,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        verify_count_remains_same: bool = False,
        mapping_file_name: Optional[str] = None,
        log_event: Optional[bool] = None,
        log_event_level: Optional[LogLevel] = None,
        log_limit: Optional[int] = None,
        desired_partitions: Optional[int] = None,
        append_to_view: Optional[bool] = None,
    ) -> None:
        """
        Runs the provided SQL and stores the result in the view parameter


        :param sql: SQL to run
        :param view: store results of running SQL into this view
        :param log_sql: whether to log the SQL to progress_logger
        :param log_result: whether to log the result to progress_logger
        :param verify_count_remains_same: whether to confirm the count of the rows remains the same after running SQL
        :param log_event: whether to log an event at TRACE level with progress_logger with the contents of
                            the output view
        :param log_event_level: the level to use for logging the event
        :param desired_partitions: (Optional) Repartition the data after executing the sql
        :param append_to_view: (Optional) Append to the view instead of overwriting it
        :param mapping_file_name: (Optional) Name of the mapping file to use for logging the result
        :param log_limit: (Optional) Limit the number of rows logged
        :param name: (Optional) Name of the transformer
        :param parameters: (Optional) Parameters for the transformer
        :param progress_logger: (Optional) ProgressLogger to use
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )
        self.logger = get_logger(__name__)

        if not sql:
            raise ValueError("sql is None or empty")

        if not view:
            raise ValueError("view is None or empty")

        self.sql: Param[Optional[str]] = Param(self, "sql", "")
        self._setDefault(sql=None)

        self.view: Param[Optional[str]] = Param(self, "view", "")
        self._setDefault(view=None)

        self.mapping_file_name: Param[Optional[str]] = Param(
            self, "mapping_file_name", ""
        )
        self._setDefault(mapping_file_name=None)

        self.log_sql: Param[bool] = Param(self, "log_sql", "")
        self._setDefault(log_sql=False)

        self.log_result: Param[bool] = Param(self, "log_result", "")
        self._setDefault(log_result=False)

        self.log_limit: Param[Optional[int]] = Param(self, "log_limit", "")
        self._setDefault(log_limit=None)

        self.verify_count_remains_same: Param[bool] = Param(
            self, "verify_count_remains_same", ""
        )
        self._setDefault(verify_count_remains_same=None)

        self.log_event: Param[Optional[bool]] = Param(self, "log_event", "")
        self._setDefault(log_event=None)

        self.log_event_level: Param[Optional[LogLevel]] = Param(
            self, "log_event_level", ""
        )
        self._setDefault(log_event_level=None)

        self.desired_partitions: Param[Optional[int]] = Param(
            self, "desired_partitions", ""
        )
        self._setDefault(desired_partitions=desired_partitions)

        self.append_to_view: Param[Optional[bool]] = Param(self, "append_to_view", "")
        self._setDefault(append_to_view=append_to_view)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        sql_text: Optional[str] = self.getSql()
        name: Optional[str] = self.getName()
        view: Optional[str] = self.getView()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        log_result: bool = self.getLogResult()
        log_event: Optional[bool] = self.getOrDefault(self.log_event)
        log_event_level: Optional[LogLevel] = self.getOrDefault(self.log_event_level)
        log_limit: Optional[int] = self.getOrDefault(self.log_limit)
        desired_partitions: Optional[int] = self.getOrDefault(self.desired_partitions)
        append_to_view: Optional[bool] = self.getOrDefault(self.append_to_view)

        assert sql_text
        with ProgressLogMetric(
            name=name or view or "", progress_logger=progress_logger
        ):
            if progress_logger and name:
                progress_logger.log_artifact(key=f"{name}.sql.txt", contents=sql_text)
                progress_logger.write_to_log(name=name, message=sql_text)
            try:
                df = df.sparkSession.sql(sql_text)
            except Exception:
                self.logger.info(f"Error in {name}")
                self.logger.info(sql_text)
                raise

            if desired_partitions is not None:
                num_partitions: int = df.rdd.getNumPartitions()
                if num_partitions != desired_partitions:
                    df = df.repartition(desired_partitions)

            if log_result:
                log_limit = 10 if log_limit is None else log_limit
                if log_limit and log_limit > 0:
                    message = (
                        "\n"
                        + (name or view or "")
                        + f" (LIMIT {log_limit})"
                        + "\n"
                        + get_pretty_data_frame(df, log_limit, sql_text)
                    )
                    self.logger.info(message)
                    if progress_logger:
                        progress_logger.write_to_log(message)
                        if log_event or log_event_level:
                            if log_event_level:
                                progress_logger.log_event(
                                    event_name=name or view or self.__class__.__name__,
                                    event_text=message,
                                    log_level=log_event_level,
                                )
                            else:
                                progress_logger.log_event(
                                    event_name=name or view or self.__class__.__name__,
                                    event_text=message,
                                )

                        progress_logger.log_artifact(
                            key=f"{name or view}.csv",
                            contents=get_data_frame_as_csv(df=df, limit=log_limit),
                        )

            if view:
                if append_to_view:
                    original_df: DataFrame = df.sparkSession.table(view)
                    df = original_df.union(df)
                # now point the view back to it
                df.createOrReplaceTempView(view)
            self.logger.info(f"GenericSqlTransformer [{name}] finished running SQL")

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getSql(self) -> Optional[str]:
        return self.getOrDefault(self.sql)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> Optional[str]:
        return self.getOrDefault(self.view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLogSql(self) -> bool:
        return self.getOrDefault(self.log_sql)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLogResult(self) -> bool:
        return self.getOrDefault(self.log_result)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getVerifyCountRemainsSame(self) -> bool:
        return self.getOrDefault(self.verify_count_remains_same)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMappingFileName(self) -> Optional[str]:
        return self.getOrDefault(self.mapping_file_name)

    def __str__(self) -> str:
        return f"{self.mapping_file_name}: view={self.view}"



from datetime import datetime
from typing import Protocol


class OnPartitionStartEventHandler(Protocol):
    async def __call__(self, *, partition_index: int) -> None:
        """
        This function is called when a new partition is started.

        :param partition_index: the id of the partition
        """
        ...


class OnPartitionCompletionEventHandler(Protocol):
    async def __call__(self, *, partition_index: int) -> None:
        """
        This function is called when a partition is completed.

        :param partition_index: the id of the partition
        """
        ...


class OnChunkStartEventHandler(Protocol):
    async def __call__(
        self,
        *,
        partition_index: int,
        chunk_index: int,
        chunk_range: range,
        chunk_start_time: datetime
    ) -> None:
        """
        This function is called when a new chunk is started.

        :param partition_index: the id of the partition
        :param chunk_index: the id of the chunk
        """
        ...


class OnChunkCompletionEventHandler(Protocol):
    async def __call__(
        self,
        *,
        partition_index: int,
        chunk_index: int,
        chunk_range: range,
        chunk_start_time: datetime,
        chunk_end_time: datetime,
        chunk_duration: float
    ) -> None:
        """
        This function is called when a chunk is completed.

        :param partition_index: the id of the partition
        :param chunk_index: the id of the chunk
        """
        ...



import json
from os import path, makedirs, environ
from pathlib import Path
from shutil import rmtree
from urllib.parse import urljoin

import pytest
from pyspark.sql import SparkSession, DataFrame

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.fhir_sender.v2.fhir_sender import FhirSender
from spark_pipeline_framework.utilities.fhir_helpers.fhir_sender_operation import (
    FhirSenderOperation,
)
from spark_pipeline_framework.utilities.fhir_server_test_context.v1.fhir_server_test_context import (
    FhirServerTestContext,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)

import requests


@pytest.mark.parametrize("run_synchronously", [True, False])
async def test_fhir_sender_put(
    spark_session: SparkSession, run_synchronously: bool
) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath("./")
    temp_folder = data_dir.joinpath("./temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    async with FhirServerTestContext(
        resource_type="Patient"
    ) as fhir_server_test_context:

        test_files_dir: Path = data_dir.joinpath("test_files/patients_put")
        response_files_dir: Path = temp_folder.joinpath("patients-response")
        parameters = {"flow_name": "Test Pipeline V2"}

        df: DataFrame = create_empty_dataframe(spark_session=spark_session)

        fhir_server_url: str = fhir_server_test_context.fhir_server_url
        auth_client_id = fhir_server_test_context.auth_client_id
        auth_client_secret = fhir_server_test_context.auth_client_secret
        auth_well_known_url = fhir_server_test_context.auth_well_known_url
        token_url = fhir_server_test_context.get_token_url()
        assert token_url
        authorization_header = fhir_server_test_context.get_authorization_header()

        environ["LOGLEVEL"] = "DEBUG"
        # Act
        with ProgressLogger() as progress_logger:
            FhirSender(
                resource="Patient",
                server_url=fhir_server_url,
                file_path=test_files_dir,
                response_path=response_files_dir,
                progress_logger=progress_logger,
                batch_size=1,
                run_synchronously=run_synchronously,
                operation=FhirSenderOperation.FHIR_OPERATION_PUT,
                parameters=parameters,
                auth_client_id=auth_client_id,
                auth_client_secret=auth_client_secret,
                auth_well_known_url=auth_well_known_url,
            ).transform(df)

        response = requests.get(
            urljoin(fhir_server_url, "Patient/00200000000"),
            headers=authorization_header,
        )
        assert response.ok, response.text
        json_text = response.text
        obj = json.loads(json_text)
        assert obj["gender"] == "male"



from pathlib import Path
from typing import Type, List

from spark_pipeline_framework.utilities.dynamic_class_loader.v1.dynamic_class_loader import (
    DynamicClassLoader,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.standardizing_vendor import (
    StandardizingVendor,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.vendors.vendor_responses.base_vendor_api_response import (
    BaseVendorApiResponse,
)


def test_dynamic_class_loader() -> None:
    print("")
    data_dir: Path = Path(__file__).parent.joinpath("./")
    standardizing_vendor_path = data_dir.joinpath(
        "../../../helix_geolocation/v2/vendors"
    )
    print(f"standardizing_vendor_path: {standardizing_vendor_path}")
    sub_classes: List[
        Type[StandardizingVendor[BaseVendorApiResponse]]
    ] = DynamicClassLoader[StandardizingVendor[BaseVendorApiResponse]](
        StandardizingVendor, standardizing_vendor_path  # type: ignore[type-abstract]
    ).find_subclasses()
    assert len(sub_classes) == 4

    assert sorted([sub_class.__name__ for sub_class in sub_classes]) == sorted(
        [
            "GeocodioStandardizingVendor",
            "MelissaStandardizingVendor",
            "MockStandardizingVendor",
            "CensusStandardizingVendor",
        ]
    )



import logging
from os import environ
from typing import Optional, List, Dict, Any, Type, Union

from helix_fhir_client_sdk.fhir_client import FhirClient
from helix_fhir_client_sdk.structures.get_access_token_result import (
    GetAccessTokenResult,
)
from helix_fhir_client_sdk.utilities.fhir_server_helpers import FhirServerHelpers

from spark_pipeline_framework.utilities.fhir_helpers.fhir_get_access_token import (
    fhir_get_access_token_async,
)
from spark_pipeline_framework.utilities.fhir_helpers.token_helper import TokenHelper


class FhirServerTestContext:
    """
    This class cleans the FHIR server before/after running the tests.
    """

    def __init__(
        self, *, resource_type: str | List[str], owner_tag: str = "bwell"
    ) -> None:
        """
        This class cleans the FHIR server before/after running the tests.

        :param resource_type: The resource type to clean.
        :param owner_tag: The owner tag to clean.
        """
        assert resource_type is not None
        assert isinstance(resource_type, str) or isinstance(resource_type, list)
        self.resource_type: str | List[str] = resource_type
        self.owner_tag: str = owner_tag
        self.fhir_server_url: str = environ["FHIR_SERVER_URL"]
        self.auth_client_id = environ["FHIR_CLIENT_ID"]
        self.auth_client_secret = environ["FHIR_CLIENT_SECRET"]
        self.auth_well_known_url = environ["AUTH_CONFIGURATION_URI"]

    async def __aenter__(self) -> "FhirServerTestContext":
        """
        Implement the async enter method so that this class can be used as an async context manager.

        """
        if isinstance(self.resource_type, list):
            for resource_type in self.resource_type:
                await FhirServerHelpers.clean_fhir_server_async(
                    resource_type=resource_type, owner_tag=self.owner_tag
                )
        else:
            await FhirServerHelpers.clean_fhir_server_async(
                resource_type=self.resource_type, owner_tag=self.owner_tag
            )
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[Union[Type[BaseException], None]],
    ) -> None:
        """
        Implement the async exit method so that this class can be used as an async context manager.

        :param exc_type: The exception type.
        :param exc_val: The exception value.
        :param exc_tb: The exception traceback.
        """
        if isinstance(self.resource_type, list):
            for resource_type in self.resource_type:
                await FhirServerHelpers.clean_fhir_server_async(
                    resource_type=resource_type, owner_tag=self.owner_tag
                )
        else:
            await FhirServerHelpers.clean_fhir_server_async(
                resource_type=self.resource_type, owner_tag=self.owner_tag
            )

    def get_authorization_header(self) -> Dict[str, Any]:
        """
        Create the authorization header using the client id and client secret.

        :return: The authorization header.
        """
        token_url: Optional[str] = TokenHelper.get_auth_server_url_from_well_known_url(
            well_known_url=self.auth_well_known_url
        )
        assert token_url
        return TokenHelper.get_authorization_header(
            client_id=self.auth_client_id,
            client_secret=self.auth_client_secret,
            token_url=token_url,
            # scope=f"access/{self.owner_tag}.* user/*.*",
            # scope=f"access/*.* user/*.*",
            scope=None,
        )

    def get_auth_server_url_from_well_known_url(self) -> Optional[str]:
        """
        Get the authorization server URL from the well known URL.

        :return: The authorization server URL.
        """
        return TokenHelper.get_auth_server_url_from_well_known_url(
            well_known_url=self.auth_well_known_url
        )

    def get_token_url(self) -> Optional[str]:
        """
        Get the token URL.

        :return: The token URL.
        """
        return self.get_auth_server_url_from_well_known_url()

    async def create_fhir_client_async(self) -> FhirClient:
        """
        Create the FHIR client.

        :return: The FHIR client.
        """
        fhir_client = FhirClient()
        fhir_client = fhir_client.url(self.fhir_server_url)
        fhir_client = fhir_client.client_credentials(
            client_id=self.auth_client_id, client_secret=self.auth_client_secret
        )
        fhir_client = fhir_client.auth_wellknown_url(self.auth_well_known_url)
        access_token_result: GetAccessTokenResult = await self.get_access_token_async()
        fhir_client = fhir_client.set_access_token(access_token_result.access_token)
        fhir_client = fhir_client.set_access_token_expiry_date(
            access_token_result.expiry_date
        )
        return fhir_client

    async def get_access_token_async(self) -> GetAccessTokenResult:
        """
        Get the access token.

        :return: The access token.
        """
        return await fhir_get_access_token_async(
            logger=logging.getLogger(__name__),
            server_url=self.fhir_server_url,
            log_level="DEBUG",
            auth_client_id=self.auth_client_id,
            auth_client_secret=self.auth_client_secret,
            auth_well_known_url=self.auth_well_known_url,
        )



import dataclasses
from typing import Union, Optional, List, Dict

from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_udf_parameters import (
    AsyncPandasUdfParameters,
)
from spark_pipeline_framework.utilities.fhir_helpers.fhir_sender_operation import (
    FhirSenderOperation,
)


@dataclasses.dataclass
class FhirSenderParameters:
    total_partitions: int
    operation: Union[FhirSenderOperation, str]
    server_url: str
    resource_name: str
    name: Optional[str]
    auth_server_url: Optional[str]
    auth_client_id: Optional[str]
    auth_client_secret: Optional[str]
    auth_login_token: Optional[str]
    auth_scopes: Optional[List[str]]
    auth_access_token: Optional[str]
    auth_well_known_url: Optional[str]
    additional_request_headers: Optional[Dict[str, str]]
    log_level: Optional[str]
    batch_size: Optional[int]
    validation_server_url: Optional[str]
    retry_count: Optional[int]
    exclude_status_codes_from_retry: Optional[List[int]]
    pandas_udf_parameters: AsyncPandasUdfParameters
    smart_merge: Optional[bool]



from typing import NamedTuple, Dict, Optional

from spark_pipeline_framework.utilities.helix_geolocation.v1.raw_address import (
    RawAddress,
)


class VendorResponse(NamedTuple):
    vendor_name: str
    response_version: str
    api_call_response: Dict[str, str]
    related_raw_address: Optional[RawAddress] = None



from typing import List, Any, Dict

import pytest
from aioresponses import aioresponses

from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.raw_address import (
    RawAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.standardized_address import (
    StandardizedAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.vendor_response import (
    VendorResponse,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.vendors.census_standardizing_vendor import (
    CensusStandardizingVendor,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.vendors.vendor_responses.census_standardizing_vendor_api_response import (
    CensusStandardizingVendorApiResponse,
)


@pytest.mark.asyncio
async def test_standardize_async_bulk() -> None:
    raw_addresses = [
        RawAddress(
            internal_id="1",
            address_id="1",
            line1="123 Test St",
            city="Test City",
            state="CA",
            zipcode="12345",
            country="US",
        ),
        RawAddress(
            internal_id="2",
            address_id="2",
            line1="456 Another St",
            city="Another City",
            state="TX",
            zipcode="67890",
            country="US",
        ),
        RawAddress(
            internal_id="3",
            address_id="3",
            line1="457 Another St",
            city="Another City",
            state="TX",
            zipcode="67890",
            country="US",
        ),
    ]

    vendor = CensusStandardizingVendor(use_bulk_api=True)

    with aioresponses() as m:
        m.post(
            "https://geocoding.geo.census.gov/geocoder/locations/addressbatch",
            status=200,
            body="""
        1, "123 Test St, Test City, CA, 12345", "Match", "Exact", "123 TEST ST, TEST CITY, CA, 12345", "-122.419416,37.774929", "123456", "L"
        2, "456 Another St, Another City, TX, 67890", "Match", "Exact", "456 ANOTHER ST, ANOTHER CITY, TX, 67890", "-97.7431,30.2672", "654321", "R"
        """,
        )

        result: List[VendorResponse[CensusStandardizingVendorApiResponse]] = (
            await vendor.standardize_async(raw_addresses)
        )

    assert len(result) == 3
    assert result[0].api_call_response.address_id == "1"
    assert result[0].api_call_response.to_dict() == {
        "addressMatches": [
            {
                "addressComponents": {
                    "city": "TEST CITY",
                    "fromAddress": None,
                    "preDirection": None,
                    "preQualifier": None,
                    "preType": None,
                    "state": "CA",
                    "streetName": "123 TEST ST",
                    "suffixDirection": None,
                    "suffixQualifier": None,
                    "suffixType": None,
                    "toAddress": None,
                    "zip": "12345",
                },
                "coordinates": {"x": -122.419416, "y": 37.774929},
                "matchedAddress": "123 TEST ST, TEST CITY, CA, 12345",
                "tigerLine": {"side": "L", "tigerLineId": "123456"},
            }
        ],
        "address_id": "1",
        "input": {
            "address": {"address": "123 Test St, Test City, CA, 12345"},
            "benchmark": {
                "benchmarkDescription": None,
                "benchmarkName": "Public_AR_Current",
                "id": None,
                "isDefault": True,
            },
        },
    }
    assert result[1].api_call_response.address_id == "2"
    assert result[2].api_call_response.address_id == "3"
    assert result[2].api_call_response.to_dict() == {
        "addressMatches": [
            {
                "addressComponents": {
                    "city": "Another City",
                    "fromAddress": None,
                    "preDirection": None,
                    "preQualifier": None,
                    "preType": None,
                    "state": "TX",
                    "streetName": None,
                    "suffixDirection": None,
                    "suffixQualifier": None,
                    "suffixType": None,
                    "toAddress": None,
                    "zip": "67890",
                },
                "coordinates": {"x": None, "y": None},
                "matchedAddress": "457 Another St, Another City TX 67890 " "US",
                "tigerLine": {"side": "L", "tigerLineId": "107644663"},
            }
        ],
        "address_id": "3",
        "input": {
            "address": {"address": "457 Another St, Another City TX 67890 US"},
            "benchmark": {
                "benchmarkDescription": "Public_AR_Current",
                "benchmarkName": "Public_AR_Current",
                "id": "4",
                "isDefault": True,
            },
        },
    }


@pytest.mark.asyncio
async def test_standardize_async_single() -> None:
    raw_addresses = [
        RawAddress(
            internal_id="1",
            address_id="1",
            line1="123 Test St",
            city="Test City",
            state="CA",
            zipcode="12345",
            country="US",
        )
    ]

    vendor = CensusStandardizingVendor(use_bulk_api=False)

    with aioresponses() as m:
        m.get(
            "https://geocoding.geo.census.gov/geocoder/locations/onelineaddress?address=123+Test+St%252C+Test+City+CA+12345+US&benchmark=Public_AR_Current&format=json&vintage=Current_Current",
            status=200,
            payload={
                "result": {
                    "input": {
                        "address": {"address": "123 Test St, Test City, CA, 12345"},
                        "benchmark": {
                            "isDefault": True,
                            "benchmarkDescription": "Public Address Ranges - Current Benchmark",
                            "id": "4",
                            "benchmarkName": "Public_AR_Current",
                        },
                    },
                    "addressMatches": [
                        {
                            "tigerLine": {"side": "L", "tigerLineId": "123456"},
                            "coordinates": {"x": -122.419416, "y": 37.774929},
                            "addressComponents": {
                                "zip": "12345",
                                "streetName": "TEST",
                                "city": "TEST CITY",
                                "state": "CA",
                                "suffixType": "ST",
                            },
                            "matchedAddress": "123 TEST ST, TEST CITY, CA, 12345",
                        }
                    ],
                }
            },
        )

        result: List[VendorResponse[CensusStandardizingVendorApiResponse]] = (
            await vendor.standardize_async(raw_addresses)
        )

    assert len(result) == 1
    assert result[0].api_call_response.address_id == "1"


@pytest.mark.asyncio
async def test_bulk_api_call_async() -> None:
    raw_addresses = [
        RawAddress(
            internal_id="1",
            address_id="1",
            line1="123 Test St",
            city="Test City",
            state="CA",
            zipcode="12345",
            country="US",
        )
    ]

    vendor = CensusStandardizingVendor(use_bulk_api=True)

    # CSV Format
    #  "ID", "Input Address", "Match Status", "Match Type","Matched Address","Coordinates","TIGER Line ID","Side"
    response_body = '"1", "123 Test St, Test City, CA, 12345", "Match", "Exact", "123 TEST ST, TEST CITY, CA, 12345", "-122.419416,37.774929", "123456", "L"'

    with aioresponses() as m:
        m.post(
            "https://geocoding.geo.census.gov/geocoder/locations/addressbatch",
            status=200,
            body=response_body,
        )

        async for response in vendor._bulk_api_call_async(raw_addresses=raw_addresses):
            assert len(response) == 1
            assert response[0].address_id == "1"


@pytest.mark.asyncio
async def test_single_api_call_async() -> None:
    raw_address = RawAddress(
        internal_id="1",
        address_id="1",
        line1="123 Test St",
        city="Test City",
        state="CA",
        zipcode="12345",
        country="US",
    )

    vendor = CensusStandardizingVendor(use_bulk_api=False)

    with aioresponses() as m:
        m.get(
            "https://geocoding.geo.census.gov/geocoder/locations/onelineaddress?address=123+Test+St%252C+Test+City+CA+12345+US&benchmark=Public_AR_Current&format=json&vintage=Current_Current",
            status=200,
            payload={
                "result": {
                    "input": {
                        "address": {"address": "123 Test St, Test City, CA, 12345"},
                        "benchmark": {
                            "isDefault": True,
                            "benchmarkDescription": "Public Address Ranges - Current Benchmark",
                            "id": "4",
                            "benchmarkName": "Public_AR_Current",
                        },
                    },
                    "addressMatches": [
                        {
                            "tigerLine": {"side": "L", "tigerLineId": "123456"},
                            "coordinates": {"x": -122.419416, "y": 37.774929},
                            "addressComponents": {
                                "zip": "12345",
                                "streetName": "TEST",
                                "city": "TEST CITY",
                                "state": "CA",
                                "suffixType": "ST",
                            },
                            "matchedAddress": "123 TEST ST, TEST CITY, CA, 12345",
                        }
                    ],
                }
            },
        )

        async for response in vendor._single_api_call_async(
            one_line_address=raw_address.to_str(), raw_address=raw_address
        ):
            assert response.address_id == "1"


def test_parse_csv_response() -> None:
    raw_addresses = [
        RawAddress(
            internal_id="1",
            address_id="1",
            line1="123 Test St",
            city="Test City",
            state="CA",
            zipcode="12345",
            country="US",
        )
    ]

    vendor = CensusStandardizingVendor()

    csv_response: Dict[str, Any] = {
        "ID": "1",
        "Input Address": "123 Test St, Test City, CA, 12345",
        "Match Status": "Match",
        "Match Type": "Exact",
        "Matched Address": "123 TEST ST, TEST CITY, CA, 12345",
        "Coordinates": "-122.419416,37.774929",
        "TIGER Line ID": "123456",
        "Side": "L",
    }

    response = vendor._parse_csv_response(csv_response, raw_addresses=raw_addresses)
    assert response.address_id == "1"
    assert response.addressMatches is not None
    assert (
        response.addressMatches[0].matchedAddress == "123 TEST ST, TEST CITY, CA, 12345"
    )


def test_parse_geolocation_response() -> None:
    response_json: Dict[str, Any] = {
        "result": {
            "input": {
                "address": {"address": "123 Test St, Test City, CA, 12345"},
                "benchmark": {
                    "isDefault": True,
                    "benchmarkDescription": "Public Address Ranges - Current Benchmark",
                    "id": "4",
                    "benchmarkName": "Public_AR_Current",
                },
            },
            "addressMatches": [
                {
                    "tigerLine": {"side": "L", "tigerLineId": "123456"},
                    "coordinates": {"x": -122.419416, "y": 37.774929},
                    "addressComponents": {
                        "zip": "12345",
                        "streetName": "TEST",
                        "city": "TEST CITY",
                        "state": "CA",
                        "suffixType": "ST",
                    },
                    "matchedAddress": "123 TEST ST, TEST CITY, CA, 12345",
                }
            ],
        }
    }

    vendor = CensusStandardizingVendor()
    response = vendor._parse_geolocation_response(
        response_json=response_json, record_id="1"
    )
    assert response is not None
    assert response.address_id == "1"
    assert response.addressMatches is not None
    assert (
        response.addressMatches[0].matchedAddress == "123 TEST ST, TEST CITY, CA, 12345"
    )


@pytest.mark.parametrize("use_bulk_api", [True, False])
async def test_real_census_standardizing_vendor(use_bulk_api: bool) -> None:
    raw_addr_obj = RawAddress(
        address_id="10",
        line1="8300 N Lamar Blvd",
        city="Austin",
        state="TX",
        zipcode="78753",
        country="US",
        line2=None,
    )

    raw_addr_obj2 = RawAddress(
        address_id="100",
        line1="4600 Silver Hill Rd",
        city="Washington",
        state="DC",
        zipcode="20233",
        country="US",
        line2=None,
    )

    raw_addr_obj3 = RawAddress(
        address_id="1000",
        line1=None,
        city=None,
        state=None,
        zipcode=None,
        country="US",
        line2=None,
    )

    # try an address with duplicate address_id
    raw_addr_obj4 = RawAddress(
        address_id="100",
        line1="7420 Market St",
        city="Wilmington",
        state="NC",
        zipcode="285466333",
        country="US",
        line2=None,
    )

    raw_addr_obj5 = RawAddress(
        address_id="10000",
        line1="",
        city="",
        state="",
        zipcode="",
        country="",
        line2="",
    )

    vendor_responses: List[
        VendorResponse[CensusStandardizingVendorApiResponse]
    ] = await CensusStandardizingVendor(use_bulk_api=use_bulk_api).standardize_async(
        [raw_addr_obj, raw_addr_obj2, raw_addr_obj3, raw_addr_obj4, raw_addr_obj5]
    )
    assert len(vendor_responses) == 5

    # test first address
    assert vendor_responses[0].vendor_name == "census"
    first_raw_response: CensusStandardizingVendorApiResponse = vendor_responses[
        0
    ].api_call_response
    first_response: StandardizedAddress | None = (
        first_raw_response.to_standardized_address(address_id="10")
    )
    # This one is not found in Census API
    assert first_response is not None
    assert first_raw_response.input is not None
    assert first_raw_response.input.address is not None
    assert first_raw_response.input.address.address is not None
    assert first_raw_response.input.address.address.startswith("8300 N Lamar Blvd")

    # test second address
    assert vendor_responses[1].vendor_name == "census"
    second_raw_response: CensusStandardizingVendorApiResponse = vendor_responses[
        1
    ].api_call_response
    second_response: StandardizedAddress | None = (
        second_raw_response.to_standardized_address(address_id="100")
    )
    assert second_response is not None
    assert second_response.address_id == "100"
    assert second_response.line1 == "4600 SILVER HILL RD"
    assert second_response.city == "WASHINGTON"
    assert second_response.state == "DC"
    assert second_response.zipcode == "20233"
    assert second_response.country == "US"
    assert second_response.latitude is not None
    assert second_response.latitude.startswith("38.84")
    assert second_response.longitude is not None
    assert second_response.longitude.startswith("-76.92")

    third_response: StandardizedAddress | None = vendor_responses[
        2
    ].api_call_response.to_standardized_address(address_id="1000")
    assert third_response is not None
    assert vendor_responses[2].api_call_response.input is not None

    # test fourth address
    fourth_response: StandardizedAddress | None = vendor_responses[
        3
    ].api_call_response.to_standardized_address(address_id="100")
    assert fourth_response is not None
    assert fourth_response.address_id == "100"
    assert fourth_response.line1 == "7420 MARKET ST"
    assert fourth_response.city == "WILMINGTON"
    assert fourth_response.state == "NC"
    assert fourth_response.zipcode == "28411"
    assert fourth_response.country == "US"
    assert fourth_response.latitude is not None
    assert fourth_response.latitude.startswith("34.2")
    assert fourth_response.longitude is not None
    assert fourth_response.longitude.startswith("-77.8")

    # test fifth address
    fifth_response: StandardizedAddress | None = vendor_responses[
        4
    ].api_call_response.to_standardized_address(address_id="10000")
    assert fifth_response is not None
    assert fifth_response.address_id == "10000"
    assert fifth_response.line1 == ""
    assert fifth_response.city == ""
    assert fifth_response.state == ""
    assert fifth_response.zipcode == ""
    assert fifth_response.country == "US"
    assert fifth_response.latitude is None
    assert fifth_response.longitude is None
    assert fifth_response.formatted_address == ""
    assert fifth_response.standardize_vendor == "census"
    assert fifth_response.county is None



from typing import Optional

from pyspark.sql import DataFrame
from pyspark.ml.param import Param
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)


class FrameworkDDLExecutor(FrameworkTransformer):
    """
    Executes a Data Definition Language (DDL) statement on a database using a JDBC connection.
    This transformer is designed to execute DDL statements (e.g., CREATE, DROP, ALTER) on a database
    connected via JDBC. It is intended to be used within a Spark pipeline.
    Parameters:
        jdbc_url (str): The JDBC URL for connecting to the database.
        ddl (str): The DDL statement to execute. Ensure that this statement is validated to prevent SQL injection.
        name (Optional[str]): An optional name for the transformer.
        progress_logger (Optional[ProgressLogger]): Optional progress logger.

    Important considerations:
    - Be cautious when executing DDL statements, as they can modify the database schema and potentially
      lead to data loss or corruption.
    """

    def __init__(
        self,
        jdbc_url: str,
        ddl: str,
        name: Optional[str] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        super().__init__(name=name, progress_logger=progress_logger)
        self.logger.info("Initializing FrameworkDDLExecutor")
        self.jdbc_url: Param[str] = Param(self, "jdbc_url", "")
        self.ddl: Param[str] = Param(self, "ddl", "")

        self._setDefault(jdbc_url=jdbc_url)
        self._setDefault(ddl=ddl)
        self.setParams(jdbc_url=jdbc_url, ddl=ddl)

    def getJdbcUrl(self) -> str:
        return self.getOrDefault(self.jdbc_url)

    def getDDL(self) -> str:
        return self.getOrDefault(self.ddl)

    def _transform(self, df: DataFrame) -> DataFrame:
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        sc = df.sql_ctx.sparkSession.sparkContext
        jvm = sc._jvm
        conn = None
        stmt = None
        ddl = self.getDDL()
        jdbc_url = self.getJdbcUrl()

        with ProgressLogMetric(
            name=self.getName() or "FrameworkDDLExecutor",
            progress_logger=progress_logger,
        ):
            try:
                conn = jvm.java.sql.DriverManager.getConnection(jdbc_url)  # type: ignore
                stmt = conn.createStatement()
                stmt.execute(ddl)
            except Exception as e:
                self.logger.error(f"Error executing DDL: {e}")
                raise RuntimeError(f"Failed to execute DDL: {e}")
            finally:
                if stmt is not None:
                    try:
                        stmt.close()
                    except Exception as close_stmt_exc:
                        self.logger.error(f"Error closing statement: {close_stmt_exc}")
                if conn is not None:
                    try:
                        conn.close()
                    except Exception as close_conn_exc:
                        self.logger.error(f"Error closing connection: {close_conn_exc}")
        self.logger.info(f"DDL Executed successfully")
        return df



from pyspark.sql.types import (
    StructType,
    StructField,
    StringType,
    BooleanType,
)


class FhirMergeResponseItemSchema:
    updated = "updated"
    created = "created"
    deleted = "deleted"
    issue = "issue"
    id_ = "id"
    uuid = "uuid"
    resourceType = "resourceType"
    sourceAssigningAuthority = "sourceAssigningAuthority"
    resource_version = "resource_version"
    message = "message"
    error = "error"
    token = "token"
    resource_json = "resource_json"

    @staticmethod
    def get_schema() -> StructType:
        """
        Returns the schema of FhirMergeResponse

        Should match to
        https://github.com/icanbwell/helix.fhir.client.sdk/blob/main/helix_fhir_client_sdk/responses/fhir_merge_response.py
        """
        response_schema = StructType(
            [
                StructField(
                    FhirMergeResponseItemSchema.created, BooleanType(), nullable=True
                ),
                StructField(
                    FhirMergeResponseItemSchema.updated, BooleanType(), nullable=True
                ),
                StructField(
                    FhirMergeResponseItemSchema.deleted, BooleanType(), nullable=True
                ),
                StructField(
                    FhirMergeResponseItemSchema.id_, StringType(), nullable=True
                ),
                StructField(
                    FhirMergeResponseItemSchema.uuid, StringType(), nullable=True
                ),
                StructField(
                    FhirMergeResponseItemSchema.resourceType,
                    StringType(),
                    nullable=True,
                ),
                StructField(
                    FhirMergeResponseItemSchema.sourceAssigningAuthority,
                    StringType(),
                    nullable=True,
                ),
                StructField(
                    FhirMergeResponseItemSchema.resource_version,
                    StringType(),
                    nullable=True,
                ),
                StructField(
                    FhirMergeResponseItemSchema.message, StringType(), nullable=True
                ),
                StructField(
                    FhirMergeResponseItemSchema.issue, StringType(), nullable=True
                ),
                StructField(
                    FhirMergeResponseItemSchema.error, StringType(), nullable=True
                ),
                StructField(
                    FhirMergeResponseItemSchema.token, StringType(), nullable=True
                ),
            ]
        )
        return response_schema



import json
from typing import Dict, List, Optional, Any, Callable
import requests
from requests.adapters import HTTPAdapter
from urllib3 import Retry

from spark_pipeline_framework.utilities.aws.config import get_ssm_config
import structlog

from spark_pipeline_framework.utilities.helix_geolocation.v1.raw_address import (
    RawAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.standardized_address import (
    StandardizedAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.standardizing_vendor import (
    StandardizingVendor,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.vendor_response import (
    VendorResponse,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.vendor_response_key_error import (
    VendorResponseKeyError,
)

logger = structlog.get_logger(__file__)


class MelissaStandardizingVendor(StandardizingVendor):
    _RESPONSE_KEY_ERROR_THRESHOLD = 2
    # number of times Melissa is allowed to send bad response until we cancel rest of requests

    def __init__(
        self,
        license_key: str = "",
        custom_api_call: Optional[Callable[[Any], Dict[str, Any]]] = None,
        version: str = "1",
        api_calls_limit: Optional[int] = None,
    ) -> None:
        """
        This class is responsible for standardizing addresses using Melissa API

        :param license_key: License key for Melissa API
        :param custom_api_call: Custom API call to Melissa API
        :param version: Version of the response
        :param api_calls_limit: Maximum number of calls to Melissa API allowed
        """
        super().__init__(version)
        self._license_key: str = license_key
        self._custom_api_call: Optional[Callable[[Any], Dict[str, Any]]] = (
            custom_api_call
        )
        self._error_counter: int = 0
        self._api_calls_limit: Optional[int] = api_calls_limit
        self._api_calls_counter: int = 0

    def standardize(
        self, raw_addresses: List[RawAddress], max_requests: int = 100
    ) -> List[VendorResponse]:
        """
        returns the vendor specific response from the vendor
        """
        vendor_specific_addresses = self._call_std_addr_api(raw_addresses=raw_addresses)
        vendor_responses: List[VendorResponse] = super()._to_vendor_response(
            vendor_response=vendor_specific_addresses,
            raw_addresses=raw_addresses,
            vendor_name=self.get_vendor_name(),
            response_version=self.get_version(),
        )
        return vendor_responses

    @staticmethod
    def vendor_specific_to_std(
        vendor_specific_addresses: List[VendorResponse],
    ) -> List[StandardizedAddress]:
        """
        Each vendor class knows how to convert its response to StdAddress
        Note: to interpret the standardize address. reference (https://www.melissa.com/quickstart-guides/global-address)
        AddressLine1-2-3 are not consistent from address to address so we infer line1 from FormattedAddress and keep
        Line2 empty for now
        """

        std_addresses = [
            StandardizedAddress(
                address_id=(
                    a.related_raw_address.get_id()
                    if a.related_raw_address
                    else a.api_call_response["RecordID"]
                ),
                line1=str(a.api_call_response["FormattedAddress"]).split(";")[0],
                city=a.api_call_response["Locality"],
                county=a.api_call_response["SubAdministrativeArea"],
                zipcode=a.api_call_response["PostalCode"],
                state=a.api_call_response["AdministrativeArea"],
                country=a.api_call_response["CountryISO3166_1_Alpha2"],
                latitude=a.api_call_response["Latitude"],
                longitude=a.api_call_response["Longitude"],
                formatted_address=a.api_call_response["FormattedAddress"],
                standardize_vendor=a.vendor_name,
            )
            for a in vendor_specific_addresses
        ]
        return std_addresses

    def _call_std_addr_api(
        self, raw_addresses: List[RawAddress]
    ) -> List[Dict[str, str]]:
        """
        Please check https://www.melissa.com/quickstart-guides/global-address for more info
        Please make sure "License Key" is available https://www.melissa.com/user/user_account.aspx
        """

        self._api_calls_counter += 1

        if self._api_calls_limit and self._api_calls_counter > self._api_calls_limit:
            logger.error(
                f"Melissa API calls limit reached. Limit: {self._api_calls_limit}"
            )
            return []

        api_server_response: Dict[str, Any] = (
            self._custom_api_call(raw_addresses)
            if self._custom_api_call
            else self._api_call(raw_addresses)
        )

        # adding vendor so we can parse it correctly after reading response from cache in the future
        try:
            vendor_specific_addresses: List[Dict[str, str]] = api_server_response[
                "Records"
            ]

        except KeyError:
            logger.exception(
                f"{self.get_vendor_name()} response does not have a Records key. Are we out of credit?"
            )
            self._error_counter += 1
            if self._error_counter > self._RESPONSE_KEY_ERROR_THRESHOLD:
                raise VendorResponseKeyError

            return []

        return vendor_specific_addresses

    def _api_call(self, raw_addresses: List[RawAddress]) -> Dict[str, Any]:
        # making the request ready
        addresses_to_lookup: List[Dict[str, str]] = [
            {
                "RecordID": a.get_id(),
                "AddressLine1": a.to_str(),
                "Country": a.to_dict()["country"],
            }
            for a in raw_addresses
        ]

        license_key = self._get_request_credentials()["license_key"]
        URL = r"https://address.melissadata.net/v3/WEB/GlobalAddress/doglobaladdress"
        json_batch_dict = {
            "TransmissionReference": "GlobalAddressBatch",
            "CustomerID": license_key,
            "Options": "",
            "Records": addresses_to_lookup,
        }
        payload = json.dumps(json_batch_dict).encode("utf-8")
        headers = {"Accept": "application/json", "Content-Type": "application/json"}
        # send the batch request
        try:
            s = requests.session()
            retries = Retry(
                total=3, backoff_factor=0.1, status_forcelist=[429, 500, 502, 503, 504]
            )

            s.mount("http://", HTTPAdapter(max_retries=retries))
            s.mount("https://", HTTPAdapter(max_retries=retries))
            response = s.post(URL, headers=headers, data=payload)
            response.raise_for_status()

            api_response: Dict[str, List[Dict[str, str]]] = json.loads(response.text)
        except Exception as e:
            logger.exception(
                f"Error connecting to {self.get_vendor_name()}", http_error=repr(e)
            )
            raise e
        return api_response

    def _get_request_credentials(self) -> Dict[str, str]:
        if self._license_key:
            license_key = self._license_key
        else:
            base_path = "/prod/helix/external/melissa/"
            c = get_ssm_config(base_path)
            license_key = c[f"{base_path}license_key_credit"]
        return {"license_key": license_key}

    @staticmethod
    def batch_request_max_size() -> int:
        return 100

    @staticmethod
    def get_vendor_name() -> str:
        return "melissa"

    @staticmethod
    def get_record_id(record: Dict[str, str]) -> str:
        return record["RecordID"]



import json
from datetime import datetime
from logging import Logger
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Callable,
    Iterator,
    AsyncGenerator,
)

import pandas as pd

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.register import register
from spark_pipeline_framework.transformers.elasticsearch_sender.v2.elasticsearch_helpers import (
    ElasticSearchHelpers,
)
from spark_pipeline_framework.transformers.elasticsearch_sender.v2.elasticsearch_result import (
    ElasticSearchResult,
)
from spark_pipeline_framework.transformers.elasticsearch_sender.v2.elasticsearch_sender_parameters import (
    ElasticSearchSenderParameters,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_udf_parameters import (
    AsyncPandasUdfParameters,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_batch_function_run_context import (
    AsyncPandasBatchFunctionRunContext,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_dataframe_udf import (
    AsyncPandasDataFrameUDF,
)
from spark_pipeline_framework.utilities.spark_partition_information.v1.spark_partition_information import (
    SparkPartitionInformation,
)


class ElasticSearchProcessor:
    @staticmethod
    async def process_chunk(
        run_context: AsyncPandasBatchFunctionRunContext,
        input_values: List[Dict[str, Any]],
        parameters: Optional[ElasticSearchSenderParameters],
        additional_parameters: Optional[Dict[str, Any]],
    ) -> AsyncGenerator[Dict[str, Any], None]:
        """
        Process a partition of data asynchronously

        :param run_context: run context
        :param input_values: input values
        :param parameters: parameters
        :param additional_parameters: additional parameters
        :return: output values
        """
        register()
        assert parameters
        logger: Logger = get_logger(
            __name__,
            level=(
                parameters.log_level if parameters and parameters.log_level else "INFO"
            ),
        )
        spark_partition_information: SparkPartitionInformation = (
            SparkPartitionInformation.from_current_task_context(
                chunk_index=run_context.chunk_index,
            )
        )
        if parameters is not None and parameters.log_level == "DEBUG":
            # ids = [input_value["id"] for input_value in input_values]
            message: str = f"ElasticSearchProcessor:process_partition"
            # Get the current time
            current_time = datetime.now()

            # Format the time to include hours, minutes, seconds, and milliseconds
            formatted_time = current_time.strftime("%H:%M:%S.%f")[:-3]
            formatted_message: str = (
                f"{formatted_time}: "
                f"{message}"
                f" | Partition: {run_context.partition_index}/{parameters.total_partitions}"
                f" | Chunk: {run_context.chunk_index}"
                f" | range: {run_context.chunk_input_range.start}-{run_context.chunk_input_range.stop}"
                f" | {spark_partition_information}"
            )
            logger.debug(formatted_message)

        count: int = 0
        try:
            full_result: Optional[ElasticSearchResult] = None
            result: ElasticSearchResult
            async for result in ElasticSearchProcessor.send_partition_to_server_async(
                partition_index=run_context.partition_index,
                parameters=parameters,
                rows=input_values,
            ):
                if result:
                    count += result.success + result.failed
                    logger.debug(
                        f"Received result"
                        f" | Partition: {run_context.partition_index}"
                        f" | Chunk: {run_context.chunk_index}"
                        f" | Successful: {result.success}"
                        f" | Failed: {result.failed}"
                    )
                    if full_result is None:
                        full_result = result
                    else:
                        full_result.append(result)
                else:
                    count += 1
                    logger.warning(
                        f"Got None result for partition {run_context.partition_index} chunk {run_context.chunk_index}"
                    )
                    error_result: ElasticSearchResult = ElasticSearchResult(
                        url=parameters.index,
                        success=0,
                        failed=1,
                        payload=[],
                        partition_index=run_context.partition_index,
                        error="Got None result",
                    )

                    if full_result is None:
                        full_result = error_result
                    else:
                        full_result.append(error_result)
            assert full_result is not None
            yield full_result.to_dict_flatten_payload()
        except Exception as e:
            logger.error(
                f"Error processing partition {run_context.partition_index} chunk {run_context.chunk_index}: {str(e)}"
            )
            # if an exception is thrown then return an error for each row
            for _ in input_values:
                count += 1
                yield {
                    "error": str(e),
                    "partition_index": run_context.partition_index,
                    "url": parameters.index,
                    "success": 0,
                    "failed": 1,
                    "payload": json.dumps(input_values),
                }
        # we actually want to error here since something strange happened
        assert count == len(input_values), f"count={count}, len={len(input_values)}"

    @staticmethod
    def get_process_partition_function(
        *, parameters: ElasticSearchSenderParameters, batch_size: int
    ) -> Callable[[Iterable[pd.DataFrame]], Iterator[pd.DataFrame]]:
        """
        Returns a function that includes the passed parameters so that function
        can be used in a pandas_udf

        :param parameters: FhirSenderParameters
        :param batch_size: batch size
        :return: pandas_udf
        """

        return AsyncPandasDataFrameUDF(
            async_func=ElasticSearchProcessor.process_chunk,  # type: ignore[arg-type]
            parameters=parameters,
            pandas_udf_parameters=AsyncPandasUdfParameters(max_chunk_size=batch_size),
        ).get_pandas_udf()

    @staticmethod
    async def send_partition_to_server_async(
        *,
        partition_index: int,
        rows: Iterable[Dict[str, Any]],
        parameters: ElasticSearchSenderParameters,
    ) -> AsyncGenerator[ElasticSearchResult, None]:
        assert parameters.index is not None
        assert isinstance(parameters.index, str)
        assert parameters.operation is not None
        assert isinstance(parameters.operation, str)
        assert parameters.doc_id_prefix is None or isinstance(
            parameters.doc_id_prefix, str
        )

        json_data_list: List[str] = [
            r["value"] for r in rows if "value" in r and r["value"] is not None
        ]
        assert isinstance(json_data_list, list)
        assert all(isinstance(j, str) for j in json_data_list)

        logger: Logger = get_logger(
            __name__,
            level=(
                parameters.log_level if parameters and parameters.log_level else "INFO"
            ),
        )

        if len(json_data_list) > 0:
            logger.info(
                f"Sending batch {partition_index}/{parameters.total_partitions} "
                f"containing {len(json_data_list)} rows "
                f"to ES Server/{parameters.index}. [{parameters.name}].."
            )
            # send to server
            response_json: ElasticSearchResult = (
                await ElasticSearchHelpers.send_json_bundle_to_elasticsearch_async(
                    json_data_list=json_data_list,
                    index=parameters.index,
                    operation=parameters.operation,
                    logger=logger,
                    doc_id_prefix=parameters.doc_id_prefix,
                    timeout=parameters.timeout or 60,
                )
            )
            response_json.partition_index = partition_index
            yield response_json
        else:
            logger.info(
                f"Batch {partition_index}/{parameters.total_partitions} is empty"
            )
            yield ElasticSearchResult(
                url=parameters.index,
                success=0,
                failed=0,
                payload=[],
                partition_index=partition_index,
                error=None,
            )



from pathlib import Path
from typing import Dict, Type, List

from spark_pipeline_framework.utilities.dynamic_class_loader.v1.dynamic_class_loader import (
    DynamicClassLoader,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.standardizing_vendor import (
    StandardizingVendor,
)


class StandardizingVendorFactory:
    vendor_class_map: Dict[str, Type[StandardizingVendor]] = {}

    @staticmethod
    def create_vendor_map() -> None:
        """
        to create a map of vendor classes by name for fast lookup
        """
        data_dir: Path = Path(__file__).parent.joinpath("./")
        standardizing_vendor_path = data_dir.joinpath("vendors")
        sub_classes: List[Type[StandardizingVendor]] = DynamicClassLoader[
            StandardizingVendor
        ](StandardizingVendor, standardizing_vendor_path).find_subclasses()

        for sub_class in sub_classes:
            StandardizingVendorFactory.vendor_class_map[sub_class.get_vendor_name()] = (
                sub_class
            )

    @staticmethod
    def get_vendor_class(vendor_name: str) -> Type[StandardizingVendor]:
        """
        find the right vendor class for the vendor name
        """
        if StandardizingVendorFactory.vendor_class_map == {}:
            StandardizingVendorFactory.create_vendor_map()

        try:
            return StandardizingVendorFactory.vendor_class_map[vendor_name]
        except KeyError:
            raise KeyError(f"No vendor Class found for {vendor_name}!")



from importlib import import_module
from typing import Any, cast

from pyspark.sql.types import StructType


def _get_schema_class_path(fhir_version: str, fhir_resource_type_name: str) -> str:
    return f"spark_fhir_schemas.{fhir_version.lower()}.resources.{fhir_resource_type_name.lower()}"


# Dynamic Module loading for the Schema classes for resources,
#   based on the given resource types available
def fhir_get_schema_class(fhir_version: str, resource_type_name: str) -> Any:
    try:
        module: Any = import_module(
            name=_get_schema_class_path(fhir_version, resource_type_name)
        )
        return getattr(module, f"{resource_type_name}Schema")
    except ModuleNotFoundError:
        raise NotImplementedError(
            f"A {fhir_version} Schema Class for {resource_type_name} resource type is not currently supported."
        )


def fhir_get_schema_for_resource(
    fhir_version: str, resource_type_name: str
) -> StructType:
    schema_class: Any = fhir_get_schema_class(fhir_version, resource_type_name)
    schema: StructType = cast(StructType, schema_class.get_schema())
    return schema



# noinspection PyProtectedMember
from pyspark.ml.functions import vector_to_array
from pyspark.sql.dataframe import DataFrame
from sparknlp import EmbeddingsFinisher, DocumentAssembler, Finisher  # type: ignore
from sparknlp.annotator import (  # type: ignore
    SentenceEmbeddings,
    WordEmbeddingsModel,
    SentenceDetector,
    Stemmer,
    ViveknSentimentModel,
    Tokenizer,
    Normalizer,
    SentimentDetector,
    UniversalSentenceEncoder,
    ClassifierDLModel,
    Lemmatizer,
    SentimentDLModel,
)
import pyspark.sql.functions as f
from pyspark.ml.feature import HashingTF, IDF, CountVectorizer
from sparknlp.pretrained import PretrainedPipeline  # type: ignore
from pyspark.ml import Pipeline

from pyspark.ml.param import Param
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from typing import Any, Dict, Optional, List
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters


class NlpTransformer(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters  # @keyword_only
    def __init__(
        self,
        column: str,
        view: str,
        binarize_tokens: bool = True,
        tfidf_n_features: Optional[int] = None,
        condense_output_columns: bool = True,
        perform_analysis: List[str] = ["all"],
        # add your parameters here
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        """

        :param column: string column of interest to perform NLP analysis
        :param view: the view to load the data into
        :param binarize_tokens: bool whether to include the arrays for binarizing labels in addition to the vectors.
            Default: True
        :param tfidf_n_features: set N most used features in text for tf-idf analysis. Default: None (use all features)
        :param condense_output_columns: remove all non-final columns used in data processing. Default True
        :param perform_analysis: the analysis options that the user wants to be performed. Here are the options
            - "all" : default parameter
            - "character_count"
            - "word_count"
            - "sentence_embeddings"
            - "sentiment"
            - "tf_idf"
            - "bag_of_words"
            - "first_word"
        :param name: a name to use when logging information about this transformer
        :param parameters: the parameter dictionary
        :param progress_logger: the progress logger


        # SETUP HELP
        Must add the following configuration to the initial spark code
        #this jar-string behaves like a list on the backend
        One may have to add other jar packages, and add them to the jar_packages string (with a "," separating)

        # jar_packages = "mysql:mysql-connector-java:X.X.X,\
        #    com.johnsnowlabs.nlp:spark-nlp_X.X:X.X.X,\
             com.someOtherJarY:X.X.X"

        # the following works as of November 2022


        jar_packages = "mysql:mysql-connector-java:8.0.24,\
            com.johnsnowlabs.nlp:spark-nlp_2.12:4.2.1"


        spark_session = SparkSession.builder \
            .appName(XXX) \
            .master("XXX") \
            .config("spark.driver.memory", "16G") \
            .config("spark.driver.maxResultSize", "0") \
            .config("spark.kryoserializer.buffer.max", "2000M") \
            .config("spark.jars.packages",
                    jar_packages
                    )
            .getOrCreate()

        Above are the recommended options by John Snow Labs.
        The following works without some of the options as the least intrusive possible code

        spark_session = SparkSession.builder \
            .appName(XXX) \
            .master("XXX") \
            .config("spark.jars.packages",
                    jar_packages
                    )
            .getOrCreate()



        # About `perform_analysis`

        For parameter "perform_analysis" the default is all the analysis will be performed, but can select custom option
        Options for parameter: `perform_analysis`
        - "all" : default parameter
        - "character_count"
        - "word_count"
        - "sentence_embeddings"
        - "sentiment"
        - "tf_idf"
        - "bag_of_words"
        - "bag_of_words_stem"
        - "first_word"

        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        # add a param

        assert column
        assert view

        self.view: Param[str] = Param(self, "view", "")
        self._setDefault(view=None)

        self.column: Param[str] = Param(self, "column", "")
        self._setDefault(column=column)

        self.binarize_tokens: Param[bool] = Param(self, "binarize_tokens", "")
        self._setDefault(binarize_tokens=binarize_tokens)

        self.perform_analysis: Param[List[str]] = Param(self, "perform_analysis", "")
        self._setDefault(perform_analysis=perform_analysis)

        self.tfidf_n_features: Param[int] = Param(self, "tfidf_n_features", "")
        self._setDefault(tfidf_n_features=tfidf_n_features)

        self.condense_output_columns: Param[bool] = Param(
            self, "condense_output_columns", ""
        )
        self._setDefault(condense_output_columns=condense_output_columns)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        """
        Performs a series of NLP feature generations  for analysis in a spark Dataframe Environment.
        Loads end result to a Spark View

        :param DataFrame df: Dataframe for NLP Transformation
        """
        # get needed parameters
        column = self.get_column()
        view = self.get_view()
        perform_analysis = self.get_perform_analysis()
        tfidf_n_features = self.get_tfidf_n_features()
        condense_output_columns = self.get_condense_output_columns()
        # if the designation is "all" or the list is empty or non-existent.
        # perform all is boolean that will compile a dataframe with every analysis for the intended column
        if "all" in perform_analysis or len(perform_analysis) == 0:
            perform_all = True
        else:
            perform_all = False
        in_col = column
        # get data from view
        #
        df_nlp: DataFrame = df.sparkSession.table(view)
        final_columns = df_nlp.columns
        explode_vector_tf = True  # need to make into function

        # FEATURE GEN
        # get string length
        out_col = "character_count"
        if out_col in perform_analysis or perform_all:
            df_nlp = df_nlp.withColumn(out_col, f.length(in_col))
            final_columns.append(out_col)

        # get word count
        out_col = "word_count"
        if out_col in perform_analysis or perform_all:
            df_nlp = df_nlp.withColumn(out_col, f.size(f.split(f.col(in_col), " ")))
            final_columns.append(out_col)
        ################
        # TOKENIZER PIPELINE
        df_nlp = df_nlp.dropna(subset=[in_col])
        # create text column with original text, which is necessary for some pretrained pipelines
        df_nlp = df_nlp.withColumn("text", df_nlp[column])
        out_col = "ntokens"
        nlp_pipeline = self.do_prepare_for_nlp(in_col, out_col)
        nlp_model = nlp_pipeline.fit(df_nlp)
        df_nlp = nlp_model.transform(df_nlp)
        final_columns.append(out_col)
        ################

        ################
        # SPARK-NLP SENTIMENTIZER
        out_col = "sentiment"
        if out_col in perform_analysis or perform_all:
            sentiment_pipeline = self.get_sentiment_vivenk(
                document_col="document",
                normalization_col="normalized",
                output_col=out_col,
            )
            sentiment_model = sentiment_pipeline.fit(df_nlp)
            df_nlp = sentiment_model.transform(df_nlp).persist()
            final_columns.append(out_col)
        """
        ################
        # SPARK-NLP Emotion Classifier
        out_col = "emotion_classifier"

        if out_col in perform_analysis or perform_all:
            begin = time.time()
            # print("Performing Analysis: {}".format(out_col))
            # emotion_pipeline = self.get_emotions_tfhub(
            #    document_col="document", output_col=out_col)
            # emotion_model = emotion_pipeline.fit(df_nlp)
            # df_nlp = emotion_model.transform(df_nlp).persist()
            # final_columns.append(out_col)
            # print("--Time Elapsed: ", round(time.time() - begin, 2))
        ############

        ################
        # SPARK-NLP SENTIMENTIZER
        out_col = "sentiment_score"
        if out_col in perform_analysis or perform_all:
            pretrained_sentiment_pipeline = self.sentiment_pretrained()
            sentiment_model = pretrained_sentiment_pipeline.fit(df_nlp)
            df_nlp = sentiment_model.transform(df_nlp)
            prediction = sentiment_model.transform(df_nlp)
            prediction.select(out_col + ".metadata").show(truncate=False)

            final_columns.append(out_col)
        """

        # get embeddings
        out_col = "sentence_embeddings"
        if out_col in perform_analysis or perform_all:
            embeddings_pipeline = self.get_sentence_embeddings_pipeline()
            embeddings_model = embeddings_pipeline.fit(df_nlp)
            df_nlp = embeddings_model.transform(df_nlp)
            final_columns.append(out_col)

        # get tfidf
        out_col = "tf_idf"
        if out_col in perform_analysis or perform_all:
            tfidf = self.get_tfidf(
                output_col=out_col, tfidf_n_features=tfidf_n_features
            )
            tfidf_model = tfidf.fit(df_nlp)
            df_nlp = tfidf_model.transform(df_nlp)
            final_columns.append(out_col)
        """
        #11/9 Commented out to assist pytest 
        # do entity recognition
        out_col = "entities"
        if out_col in perform_analysis or "all" in perform_analysis:
            ner = self.get_entities()
            # ner_model = ner.fit(df_nlp)
            df_nlp = ner.transform(df_nlp)
            final_columns.append(out_col)
        """
        # label binarizer
        # get bag of words
        out_col = "bag_of_words"
        if out_col in perform_analysis or perform_all:
            explode_col = "binarize_bag_of_words"
            df_nlp = self.do_bag_of_words(
                df=df_nlp,
                tokens_col="ntokens",
                out_col=out_col,
                explode_vector=explode_vector_tf,
                explode_col=explode_col,
            )

            final_columns.append(out_col)
            final_columns.append(explode_col)

        out_col = "bag_of_words_stem"
        if out_col in perform_analysis or perform_all:
            explode_col = "binarize_bag_of_words_stem"
            df_nlp = self.do_bag_of_words(
                df=df_nlp,
                tokens_col="ntokens_stem",
                out_col=out_col,
                explode_vector=explode_vector_tf,
                explode_col=explode_col,
            )

            final_columns.append(out_col)
            final_columns.append(explode_col)
        # first word binarizer

        # first word
        out_col = "first_word"
        if out_col in perform_analysis or perform_all:
            out_col_fw = "first_word0"
            out_col_fwl = out_col_fw + "_L"
            in_col_t = "ntokens"

            # get first word from token list
            df_nlp = df_nlp.withColumn(out_col_fw, df_nlp[in_col_t].getItem(0))
            # convert to a list, then it will work with our bag of words codes
            df_nlp = df_nlp.withColumn(out_col_fwl, f.split(out_col_fw, ","))
            explode_col = "binarize_first_word"
            df_nlp = self.do_bag_of_words(
                df=df_nlp,
                tokens_col=out_col_fwl,
                out_col=out_col,
                explode_vector=explode_vector_tf,
                explode_col=explode_col,
            )
            final_columns.append(out_col)
            final_columns.append(explode_col)

        ##################
        # finalizing output
        if condense_output_columns:
            print("##########")
            print("Condensing Output")
            df_nlp = df_nlp.select(final_columns)
            print("##########")

        df_nlp.createOrReplaceTempView(view)  # loading to view
        # df_result.createOrReplaceTempView(view)
        return df_nlp

    def do_prepare_for_nlp(
        self, input_col: str = "text", output_col: str = "ntokens"
    ) -> Pipeline:
        """
        Preforms a multistep process to generate many of the column entities needed for NLP Analysis.
        DocumentAssembler--> SentenceDetector--> Tokenizer --> Stemmer--> Normalizer--> Finisher

        :param str input_col: Text column that is the target of the analysis
        :param str output_col: Name of finished token output column

        :rtype: Pipeline
        """

        pipeline_list = []
        document_assembler = (
            DocumentAssembler().setInputCol(input_col).setOutputCol("document")
        )
        pipeline_list.append(document_assembler)

        sentence_detector = (
            SentenceDetector()
            .setInputCols(["document"])
            .setOutputCol("sentence")
            .setUseAbbreviations(True)
        )
        pipeline_list.append(sentence_detector)

        tokenizer = Tokenizer().setInputCols(["sentence"]).setOutputCol("token")
        pipeline_list.append(tokenizer)

        stemmer = Stemmer().setInputCols(["token"]).setOutputCol("stem")
        pipeline_list.append(stemmer)

        normalizer = Normalizer().setInputCols(["stem"]).setOutputCol("normalized")
        pipeline_list.append(normalizer)

        normalizer_token = (
            Normalizer()
            .setInputCols(["token"])
            .setOutputCol("normalized_token")
            .setLowercase(True)
        )
        pipeline_list.append(normalizer_token)

        output_col_stem = output_col + "_stem"
        finisher = (
            Finisher()
            .setInputCols(["normalized"])
            .setOutputCols([output_col_stem])
            .setOutputAsArray(True)
            .setCleanAnnotations(False)
        )
        pipeline_list.append(finisher)

        finished_token = (
            Finisher()
            .setInputCols(["normalized_token"])
            .setOutputCols([output_col])
            .setOutputAsArray(True)
            .setCleanAnnotations(False)
        )
        pipeline_list.append(finished_token)

        nlp_pipeline = Pipeline(stages=pipeline_list)
        return nlp_pipeline

    def get_sentiment_vivenk(
        self,
        document_col: str = "document",
        normalization_col: str = "normalized",
        output_col: str = "final_sentiment",
    ) -> Pipeline:
        """
        Generates a pipeline that creates sentiment values from documents and normalization columns

        :param str document_col: Column with documents typically created with documentAssembler()
        :param str normalization_col: Name of input column with tokens typically created with Normalizer()
        :param str output_col: Name of sentiment output column

        :rtype: Pipeline
        """
        vivekn = (
            ViveknSentimentModel.pretrained()
            .setInputCols([document_col, normalization_col])
            .setOutputCol("result_sentiment")
        )
        finisher = (
            Finisher()
            .setInputCols(["result_sentiment"])
            .setOutputCols(output_col)
            .setCleanAnnotations(False)
        )

        pipeline = Pipeline().setStages([vivekn, finisher])
        return pipeline

    """
    def sentiment_pretrained(
            self,
            document_col: str = "documents",
            output_col: str = "sentiment_score",
            sentiment_model: str = "analyze_sentimentdl_glove_imdb",
    ) -> PretrainedPipeline:
        #""
        Performs Sentiment Analysis
        :param str document_col: Column with documents typically created with documentAssembler()
        :param str output_col: Name of sentiment output column
        :param str sentiment_model: The pretrained model that is used for analysis.
        :rtype: PretrainedPipeline

        Possible configurations for sentiment model:
         -- "analyze_sentiment"
         -- "analyze_sentimentdl_glove_imdb"
         -- "analyze_sentimentdl_use_twitter"
         -- "sentimentdl_use_imdb"
         -- "sentimentdl_use_twitter"
        #""
        use = (
            UniversalSentenceEncoder.pretrained()
            .setInputCols([document_col])
            .setOutputCol("sentence_embeddings")
        )

        sentimentdl = (
            SentimentDLModel.load("analyze_sentimentdl_use_twitter")
            .setInputCols(["sentence_embeddings"])
            .setOutputCol(output_col)
        )

        nlpPipeline = Pipeline(stages=[use, sentimentdl])
        # example
        #        # pipeline = PretrainedPipeline("onto_recognize_entities_sm")

        # pipeline = PretrainedPipeline(sentiment_model)
        return nlpPipeline
    """

    def get_sentence_embeddings_pipeline(
        self,
        document_col: str = "document",
        token_col: str = "token",
        output_col: str = "sentence_embeddings",
    ) -> Pipeline:
        """
        Generates a pipeline that creates sentence embeddings transformer

        :param str document_col: Column with documents typically created with documentAssembler()
        :param str token_col: Name of input column with tokens typically created with Tokenizer()
        :param str output_col: Name of embeddings output column

        :rtype: Pipeline
        """
        pipeline_list = []
        embeddings = (
            WordEmbeddingsModel.pretrained()
            .setInputCols(document_col, token_col)
            .setOutputCol("embeddings0")
        )
        pipeline_list.append(embeddings)

        embeddings_sentence = (
            SentenceEmbeddings()
            .setInputCols([document_col, "embeddings0"])
            .setOutputCol("sentence_embeddings0")
            .setPoolingStrategy("AVERAGE")
        )
        pipeline_list.append(embeddings_sentence)

        # puts embeddings into an analysis ready form
        embeddings_finisher = (
            EmbeddingsFinisher()
            .setInputCols("sentence_embeddings0")
            .setOutputCols(output_col)
            .setOutputAsVector(True)
            .setCleanAnnotations(False)
        )
        pipeline_list.append(embeddings_finisher)

        nlp_pipeline = Pipeline(
            stages=pipeline_list
        )  # [embeddings, embeddingsSentence, embeddingsFinisher]
        return nlp_pipeline

    def do_bag_of_words(
        self,
        df: DataFrame,
        tokens_col: str = "ntokens",
        out_col: str = "bag_of_words_vectors",
        explode_col: str = "binarized_bag_of_words",
        explode_vector: bool = True,
    ) -> DataFrame:
        """
        Generates a transformer that creates a bag of words analysis from a list of tokens

        :param DataFrame df: The person sending the message
        :param str tokens_col: Column with array of tokens
        :param str out_col: Name of vector output column
        :param str explode_col: Name of exploded vector column
        :param bool explode_vector: Whether to create the exploded column
        :rtype: DataFrame
        """

        cv = CountVectorizer()
        cv.setInputCol(tokens_col)
        cv.setOutputCol(out_col)
        model = cv.fit(df)
        df = model.transform(df)
        if explode_vector:
            df = df.withColumn(
                explode_col, vector_to_array(df[out_col])
            )  # arg-type: ignore

        return df

    def get_tfidf(
        self,
        token_col: str = "ntokens",
        output_col: str = "tf_idf_features",
        tfidf_n_features: Optional[int] = None,
    ) -> Pipeline:
        """
        Generates a tf-idf analysis.

        :param str token_col: Column with array of tokens
        :param str output_col: Name of vector output column
        :param str tfidf_n_features: set N most used features in text for tf-idf analysis. Default: use all features
        :rtype: Pipeline
        """
        hashing_output_name = "tf_idf_raw_features"
        tf_hash = HashingTF().setInputCol(token_col).setOutputCol(hashing_output_name)
        if tfidf_n_features:
            tf_hash.setNumFeatures(tfidf_n_features)
        idf = IDF().setInputCol(hashing_output_name).setOutputCol(output_col)
        tfidf_pipeline = Pipeline(stages=[tf_hash, idf])

        return tfidf_pipeline

    """
    def get_entities(
        self, entity_pipeline: str = "onto_recognize_entities_sm"
    ) -> PretrainedPipeline:

        ""#"
        Creates transformer for pretrained entity recognition pipeline. In a Spark DataFrame environment.
        Code looks for "text" column
        Options for `entity_pipeline` are:
        - "onto_recognize_entities_sm"
        - "onto_recognize_entities_bert_tiny"

        :param str entity_pipeline: The name of the pretrained pipeline to use
        :rtype PretrainedPipeline
        ""#"
        # pipeline = PretrainedPipeline("onto_recognize_entities_sm")

        pipeline = PretrainedPipeline(entity_pipeline)

        return pipeline
    """

    def get_emotions_tfhub(
        self, document_col: str = "document", output_col: str = "emotion_classifier"
    ) -> Pipeline:
        """
        Can classify various emotions in text.
        - Surprise
        - Sadness
        - Fear
        - Joy

        :param str document_col: contains the DocumentAssembler column name
        :param str output_col: Name of embeddings output column
        :rtype Pipeline
        """
        universal_sentence_encoder = (
            UniversalSentenceEncoder.pretrained("tfhub_use", lang="en")
            .setInputCols([document_col])
            .setOutputCol("sentence_embeddings")
        )

        emotion_classifier = (
            ClassifierDLModel.pretrained("classifierdl_use_emotion", "en")
            .setInputCols([document_col, "sentence_embeddings"])
            .setOutputCol(output_col)
        )

        emotion_classifier_pipeline = Pipeline(
            stages=[universal_sentence_encoder, emotion_classifier]
        )

        return emotion_classifier_pipeline

    # parameter getters
    def get_column(self) -> str:
        return self.getOrDefault(self.column)

    def get_view(self) -> str:
        return self.getOrDefault(self.view)

    def get_binarize_tokens(self) -> bool:
        return self.getOrDefault(self.binarize_tokens)

    def get_perform_analysis(self) -> List[str]:
        return self.getOrDefault(self.perform_analysis)

    def get_tfidf_n_features(self) -> int:
        return self.getOrDefault(self.tfidf_n_features)

    def get_condense_output_columns(self) -> bool:
        return self.getOrDefault(self.condense_output_columns)



from typing import List, Optional, Any, Dict
from pyspark.sql.types import Row


class FhirGetResponseWriter:
    @staticmethod
    def create_row(
        partition_index: int,
        sent: int,
        received: int,
        responses: List[str],
        first: Optional[str],
        last: Optional[str],
        error_text: Optional[str],
        url: Optional[str],
        status_code: int,
        request_id: Optional[str],
        extra_context_to_return: Optional[Dict[str, Any]],
        access_token: Optional[str],
    ) -> Row:
        return Row(
            partition_index=partition_index,
            sent=sent,
            received=received,
            responses=responses,
            first=first,
            last=last,
            error_text=error_text,
            url=url,
            status_code=status_code,
            request_id=request_id,
            extra_context_to_return=extra_context_to_return,
            access_token=access_token,
        )



from datetime import datetime

import pytest
from aioresponses import aioresponses

from spark_pipeline_framework.transformers.elasticsearch_sender.v2.elasticsearch_processor import (
    ElasticSearchProcessor,
)
from spark_pipeline_framework.transformers.elasticsearch_sender.v2.elasticsearch_sender_parameters import (
    ElasticSearchSenderParameters,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_batch_function_run_context import (
    AsyncPandasBatchFunctionRunContext,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_udf_parameters import (
    AsyncPandasUdfParameters,
)


@pytest.mark.asyncio
async def test_process_partition_success() -> None:
    parameters = ElasticSearchSenderParameters(
        index="test_index",
        operation="index",
        log_level="DEBUG",
        total_partitions=1,
        name="test_name",
        doc_id_prefix=None,
        timeout=60,
        pandas_udf_parameters=AsyncPandasUdfParameters(),
    )
    input_values = [{"value": '{"key": "value"}'}]
    chunk_input_range = range(0, 1)

    with aioresponses() as m:
        m.post(
            "https://elasticsearch:9200/test_index/_bulk",
            payload={
                "items": [{"index": {"_id": "1", "status": 201}}],
                "errors": False,
            },
        )

        async_gen = ElasticSearchProcessor.process_chunk(
            run_context=AsyncPandasBatchFunctionRunContext(
                partition_index=0,
                chunk_index=0,
                chunk_input_range=chunk_input_range,
                partition_start_time=datetime.now(),
            ),
            input_values=input_values,
            parameters=parameters,
            additional_parameters=None,
        )

        results = [result async for result in async_gen]
        assert len(results) == 1, f"results: {results}"
        assert results[0]["success"] == 1, f"results: {results}"
        assert results[0]["failed"] == 0, f"results: {results}"


@pytest.mark.asyncio
async def test_process_partition_failure() -> None:
    parameters = ElasticSearchSenderParameters(
        index="test_index",
        operation="index",
        log_level="DEBUG",
        total_partitions=1,
        name="test_name",
        doc_id_prefix=None,
        timeout=60,
        pandas_udf_parameters=AsyncPandasUdfParameters(),
    )
    input_values = [{"value": '{"key": "value"}'}]
    chunk_input_range = range(0, 1)

    with aioresponses() as m:
        m.post("https://elasticsearch:9200/test_index/_bulk", status=500)

        async_gen = ElasticSearchProcessor.process_chunk(
            run_context=AsyncPandasBatchFunctionRunContext(
                partition_index=0,
                chunk_index=0,
                chunk_input_range=chunk_input_range,
                partition_start_time=datetime.now(),
            ),
            input_values=input_values,
            parameters=parameters,
            additional_parameters=None,
        )

        results = [result async for result in async_gen]
        assert len(results) == 1, f"results: {results}"
        assert results[0]["success"] == 0, f"results: {results}"
        assert results[0]["failed"] == 1, f"results: {results}"


@pytest.mark.asyncio
async def test_send_partition_to_server_async() -> None:
    parameters = ElasticSearchSenderParameters(
        index="test_index",
        operation="index",
        log_level="DEBUG",
        total_partitions=1,
        name="test_name",
        doc_id_prefix=None,
        timeout=60,
        pandas_udf_parameters=AsyncPandasUdfParameters(),
    )
    rows = [{"value": '{"key": "value"}'}]

    with aioresponses() as m:
        m.post(
            "https://elasticsearch:9200/test_index/_bulk",
            payload={
                "items": [{"index": {"_id": "1", "status": 201}}],
                "errors": False,
            },
        )

        async_gen = ElasticSearchProcessor.send_partition_to_server_async(
            partition_index=0, rows=rows, parameters=parameters
        )

        results = [result async for result in async_gen]
        assert len(results) == 1, f"results: {results}"
        assert results[0].success == 1, f"results: {results}"
        assert results[0].failed == 0, f"results: {results}"



import os
import traceback
from typing import Any, Optional, List, Dict

from py4j.protocol import Py4JJavaError
from pyspark.errors import AnalysisException

from spark_pipeline_framework.transformers.framework_mapping_runner.v1.framework_mapping_runner_exception import (
    FrameworkMappingRunnerException,
)
from spark_pipeline_framework.utilities.cannot_cast_exception_parser import (
    parse_cannot_cast_exception,
)


class FriendlySparkException(Exception):
    """
    A comprehensive exception handler for Spark-related errors.

    Provides detailed error context, traceback, and nested exception information.
    """

    def __init__(
        self,
        exception: Exception,
        stage_name: Optional[str] = None,
        message: Optional[str] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a friendly Spark exception with comprehensive error details.

        :param exception: The original exception
        :param stage_name: Name of the stage where the exception occurred
        :param message: Additional error message
        :param args: Additional positional arguments
        :param kwargs: Additional keyword arguments
        """
        # Initialize base error attributes
        self.original_exception: Exception = exception
        self.stage_name: Optional[str] = stage_name

        # Construct detailed error message
        error_message = self._construct_error_message(exception, stage_name, message)

        # Call parent exception constructor
        super().__init__(error_message)

    def _construct_error_message(
        self, exception: Exception, stage_name: Optional[str], message: Optional[str]
    ) -> str:
        """
        Construct a comprehensive error message.

        :param exception: The original exception
        :param stage_name: Name of the stage where the exception occurred
        :param message: Additional error message
        :return: Formatted error message
        """
        # Start with optional custom message
        full_message = f"{message}\n" if message else ""

        # Handle specific exception types
        exception_type = type(exception)
        exception_handlers = {
            AnalysisException: str,
            FrameworkMappingRunnerException: str,
            Py4JJavaError: str,
            Exception: str,
        }

        # Use specific handler or default to str()
        handler = exception_handlers.get(exception_type, str)
        full_message += handler(exception)

        # Add stage name if available
        if stage_name:
            full_message += f"\nStage: {stage_name}"

        # Add full traceback
        full_message += "\n" + self._get_exception_traceback(exception)

        return full_message

    def _get_exception_traceback(self, exception: Exception) -> str:
        """
        Generate a formatted traceback for the given exception.

        :param exception: The exception to generate traceback for
        :return: Formatted traceback string
        """
        return "".join(traceback.TracebackException.from_exception(exception).format())

    @classmethod
    def parse_nested_exceptions(cls, exception: Exception) -> List[str]:
        """
        Parse nested exceptions to extract detailed error information.

        :param exception: The root exception
        :return: List of parsed exception messages
        """
        nested_messages = []
        current_exception: Optional[BaseException] = exception

        while current_exception:
            # Extract exception details
            file_name = cls._get_exception_file_name(current_exception)
            line_number = cls._get_exception_line_number(current_exception)

            # Handle specific parsing for cast exceptions
            exception_message = str(current_exception)
            if "cannot cast " in exception_message:
                cast_parse_result = parse_cannot_cast_exception(exception_message)
                nested_messages.extend(
                    [
                        f"{item} [{file_name} {'(' + str(line_number) + ')' if line_number else ''}]"
                        for item in cast_parse_result
                    ]
                )

            # Add general exception message
            nested_messages.append(
                f"{exception_message} [{file_name} {'(' + str(line_number) + ')' if line_number else ''}]"
            )

            # Move to next nested exception
            current_exception = getattr(current_exception, "__cause__", None)

        return nested_messages

    @staticmethod
    def _get_exception_file_name(exception: BaseException) -> str:
        """
        Get the filename where the exception occurred.

        :param exception: The exception
        :return: Filename or empty string
        """
        if hasattr(exception, "__traceback__") and exception.__traceback__:
            return os.path.basename(exception.__traceback__.tb_frame.f_code.co_filename)
        return ""

    @staticmethod
    def _get_exception_line_number(exception: BaseException) -> Optional[int]:
        """
        Get the line number where the exception occurred.

        :param exception: The exception
        :return: Line number or None
        """
        if hasattr(exception, "__traceback__") and exception.__traceback__:
            return exception.__traceback__.tb_frame.f_code.co_firstlineno
        return None

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert exception details to a dictionary for logging.

        :return: Dictionary of exception details
        """
        return {
            "message": str(self),
            "stage_name": self.stage_name,
            "exception_type": type(self.original_exception).__name__,
            "nested_exceptions": self.parse_nested_exceptions(self.original_exception),
        }



import pytest
from aioresponses import aioresponses
from typing import List, Optional
from spark_pipeline_framework.utilities.fhir_helpers.fhir_parse_bundles import (
    combine_bundles,
    extract_resource_from_json,
)

# Sample JSON data for testing
sample_json = """
[
    {
        "resourceType": "Patient",
        "id": "example"
    },
    {
        "resourceType": "Bundle",
        "entry": [
            {
                "resource": {
                    "resourceType": "Observation",
                    "id": "obs1"
                }
            },
            {
                "resource": {
                    "resourceType": "Observation",
                    "id": "obs2"
                }
            }
        ]
    }
]
"""


def test_combine_bundles() -> None:
    """
    Test the combine_bundles function
    """
    result = combine_bundles(sample_json)
    assert result["resourceType"] == "Bundle"
    assert result["type"] == "collection"
    assert len(result["entry"]) == 3
    assert result["entry"][0]["resource"]["resourceType"] == "Patient"
    assert result["entry"][1]["resource"]["resourceType"] == "Observation"
    assert result["entry"][2]["resource"]["resourceType"] == "Observation"


def test_extract_resource_from_json() -> None:
    """
    Test the extract_resource_from_json function
    """
    file_path = "test_file.json"
    resources_to_extract: Optional[List[str]] = ["Patient", "Observation"]
    result = list(
        extract_resource_from_json(file_path, sample_json, resources_to_extract)
    )

    assert len(result) == 3
    assert result[0]["resourceType"] == "Patient"
    assert result[1]["resourceType"] == "Observation"
    assert result[2]["resourceType"] == "Observation"


@pytest.mark.asyncio
async def test_extract_resource_from_json_with_http_mock() -> None:
    """
    Test the extract_resource_from_json function with mocked HTTP requests
    """
    with aioresponses() as m:
        # Mock HTTP requests if needed
        # Example: m.get('http://example.com/api/resource', payload={'key': 'value'})

        file_path = "test_file.json"
        resources_to_extract: Optional[List[str]] = ["Patient", "Observation"]
        result = list(
            extract_resource_from_json(file_path, sample_json, resources_to_extract)
        )

        assert len(result) == 3
        assert result[0]["resourceType"] == "Patient"
        assert result[1]["resourceType"] == "Observation"
        assert result[2]["resourceType"] == "Observation"



from typing import Dict, Any, Optional, List, Set

# noinspection PyProtectedMember
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.sql.types import Row, ArrayType, StringType
from pyspark.ml.param import Param
from pyspark.sql import Column
from pyspark.sql.dataframe import DataFrame

# noinspection PyUnresolvedReferences
from pyspark.sql.functions import (
    col,
    from_json,
    get_json_object,
    transform,
    array_distinct,
)
from pyspark.sql.functions import filter
from spark_fhir_schemas.r4.resources.resource_schema_index import ResourceSchemaIndex
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class FhirBundleSplitter(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        view: str,
        fail_fast: bool = True,
        resource_types: Optional[List[str]] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        self.view: Param[str] = Param(self, "view", "")
        self._setDefault(view=view)

        self.fail_fast: Param[bool] = Param(self, "fail_fast", "")
        self._setDefault(fail_fast=fail_fast)

        self.resource_types: Param[Optional[List[str]]] = Param(
            self, "resource_types", ""
        )
        self._setDefault(resource_types=resource_types)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view: str = self.getView()
        fail_fast: bool = self.getFailFast()
        resource_types: Optional[List[str]] = self.getResourceTypes()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()

        # Assumes the schema to be BundleSchema.get_schema()
        df = df.sparkSession.table(view)

        # if caller did not pass resource types then get them from the data (albeit pretty slow)
        if resource_types is None:
            if progress_logger:
                progress_logger.write_to_log(
                    "WARNING: No resource types are passed. Auto detection in progress... This can be slow."
                )
            resource_types_rows: List[Row] = df.select(
                array_distinct(
                    transform(
                        col("entry"),
                        lambda x: get_json_object(x["resource"], "$.resourceType"),
                    )
                )
            ).collect()

            resource_types_set: Set[str] = set()
            for resource_types_row in resource_types_rows:
                for resource_types_list in resource_types_row:
                    for resource_type in resource_types_list:
                        resource_types_set.add(resource_type)
            resource_types = list(resource_types_set)

            if progress_logger:
                progress_logger.write_to_log(
                    "Following resource types were found in the data:"
                )
                progress_logger.write_to_log(",".join(resource_types))

        column_specs: Dict[str, Column] = {
            resource_type.lower(): from_json(
                filter(
                    df["entry"],
                    lambda x: get_json_object(x["resource"], "$.resourceType")
                    == resource_type,
                )
                .getField("resource")
                .cast(StringType()),
                schema=ArrayType(ResourceSchemaIndex.get(resource_type).get_schema()),
                options={"mode": "FAILFAST"} if fail_fast else None,
            )
            for resource_type in resource_types
        }

        df = df.withColumns(column_specs)
        df.createOrReplaceTempView(view)
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFailFast(self) -> bool:
        return self.getOrDefault(self.fail_fast)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getResourceTypes(self) -> Optional[List[str]]:
        return self.getOrDefault(self.resource_types)



from os import environ
from typing import Iterable, Any, Generator, Tuple, List

from pyspark.sql import SparkSession, DataFrame

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_partitioner.v2.framework_partitioner import (
    FrameworkPartitioner,
)


def test_framework_partitioner_calculate_automatically(
    spark_session: SparkSession,
) -> None:
    print()
    df: DataFrame = spark_session.createDataFrame(
        [
            ("00100000000", "Qureshi"),
            ("00200000000", "Vidal"),
            ("00300000000", "Smith"),
            ("00400000000", "Johnson"),
            ("00500000000", "Williams"),
            ("00600000000", "Brown"),
            ("00700000000", "Jones"),
            ("00800000000", "Garcia"),
            ("00900000000", "Miller"),
            ("01000000000", "Davis"),
            ("01100000000", "Rodriguez"),
            ("01200000000", "Martinez"),
            ("01300000000", "Hernandez"),
            ("01400000000", "Lopez"),
            ("01500000000", "Gonzalez"),
            ("01600000000", "Wilson"),
            ("01700000000", "Anderson"),
            ("01800000000", "Thomas"),
            ("01900000000", "Taylor"),
            ("02000000000", "Moore"),
        ],
        ["id", "name"],
    )

    df.createOrReplaceTempView("test_view")

    print(f"Partition Count: {df.rdd.getNumPartitions()}")

    # Function to list partition IDs and column values
    def list_partition_ids_and_values(
        partition_index: int, iterator: Iterable[Any]
    ) -> Generator[Tuple[int, List[str]], None, None]:
        for row in iterator:
            yield partition_index, row["id"]

    partition_id_and_values: List[Tuple[int, List[str]]] = (
        df.rdd.mapPartitionsWithIndex(list_partition_ids_and_values).collect()
    )
    print(f"Partition IDs: {partition_id_and_values}")

    environ["LOGLEVEL"] = "DEBUG"
    # Act
    with ProgressLogger() as progress_logger:
        result_df = FrameworkPartitioner(
            view="test_view",
            enable_repartitioning=True,
            calculate_automatically=True,
            name="FrameworkPartitioner",
            progress_logger=progress_logger,
        ).transform(df)

    # Assert
    # we should get two partitions since there are two executors and the memory available to executors is 2g
    assert result_df.rdd.getNumPartitions() == 2
    assert result_df.count() == 20
    assert result_df.columns == ["id", "name"]



import os
import re
from pathlib import Path
from typing import Any, Dict, Optional, Union

# noinspection PyProtectedMember
from urllib.parse import urlparse, ParseResult
from zipfile import is_zipfile, ZipFile

import boto3
import requests
from boto3.s3.transfer import TransferConfig

from spark_pipeline_framework.utilities.api_helper.http_request import (
    HelixHttpRequest,
    RequestType,
)
from spark_pipeline_framework.utilities.aws.s3 import parse_s3_uri
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from requests.adapters import HTTPAdapter
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from urllib3 import Retry


class DownloadFileTransformer(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        source_url: str,
        download_to_path: Union[str, Path],
        file_name_map: Dict[str, str],
        s3_destination_path: str,
        extract_zip: Optional[bool] = False,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        """
        Transformer to download files from a source url and upload them to s3
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        self.source_url: Param[str] = Param(self, "source_url", "")
        self._setDefault(source_url=source_url)

        self.download_to_path: Param[str] = Param(self, "download_to_path", "")
        self._setDefault(download_to_path=download_to_path)

        self.extract_zip: Param[bool] = Param(self, "extract_zip", "")
        self._setDefault(extract_zip=extract_zip)

        self.s3_destination_path: Param[str] = Param(self, "s3_destination_path", "")
        self._setDefault(s3_destination_path=s3_destination_path)

        self.file_name_map: Param[Dict[str, str]] = Param(self, "file_name_map", "")
        self._setDefault(file_name_map=file_name_map)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @staticmethod
    def check_if_path_exists(path: str) -> bool:
        if not os.path.exists(path):
            raise FileNotFoundError(f"File does not exists at this path: {path}")
        return True

    def _transform(self, df: DataFrame) -> DataFrame:
        url: str = self.getSourceUrl()
        download_path: str = self.getDownloadToPath()
        extract_zip: bool = self.getExtractZip()
        s3_destination_path: str = self.getS3DestinationPath()
        file_pattern_map: Dict[str, str] = self.GetFileNameMap()
        name: Optional[str] = self.getName()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        s3_client = boto3.client("s3")

        with ProgressLogMetric(name=name or url, progress_logger=progress_logger):
            url_parse_result: ParseResult = urlparse(url)
            downloaded_file_path: str = os.path.join(
                download_path, os.path.basename(url_parse_result.path)
            )

            session = requests.session()
            retries = Retry(
                total=3,
                backoff_factor=0.1,
                status_forcelist=[429, 500, 502, 503, 504],
            )
            session.mount("http://", HTTPAdapter(max_retries=retries))
            session.mount("https://", HTTPAdapter(max_retries=retries))

            self.logger.info(f"making HEAD request to url: {url})")

            # need to confirm the data exists at the url specified
            helix_request = HelixHttpRequest(
                url=url, request_type=RequestType.HEAD
            ).get_response()

            if helix_request.status_code == 404:
                self.logger.info(
                    f"received 404 response from {url}. nothing to download"
                )
            else:
                self.logger.info(
                    f"data is available. starting the download to {downloaded_file_path}"
                )
                try:
                    with session.get(url, stream=True, allow_redirects=True) as r:
                        r.raise_for_status()
                        with open(downloaded_file_path, "wb") as f:
                            for chunk in r.iter_content(chunk_size=8192):
                                if chunk:
                                    f.write(chunk)
                except requests.exceptions.RequestException as e:
                    self.logger.exception(
                        f"there was an error downloading data from: {url}. error message: {e}"
                    )
                    raise Exception(
                        f"there was an error downloading data from: {url}. error message: {e}"
                    )

                if extract_zip:
                    self.logger.info("download complete. beginning extract archive")
                    if self.check_if_path_exists(downloaded_file_path):
                        if is_zipfile(downloaded_file_path):
                            with ZipFile(downloaded_file_path, "r") as zip_ref:
                                zip_ref.extractall(download_path)

                s3_bucket, s3_prefix = parse_s3_uri(s3_destination_path)

                if (s3_bucket is None) or (s3_prefix is None):
                    raise Exception(
                        f"invalid s3 path provided! attempted to s3 url: {s3_destination_path}"
                    )

                for dest_file_name, file_pattern in file_pattern_map.items():
                    files = [
                        f
                        for f in os.listdir(download_path)
                        if re.match(file_pattern, f)
                    ]

                    if len(files) == 0:
                        self.logger.info(f"no files matched pattern: {file_pattern}")

                    # we only support uploading a single file per file pattern so error if more than one is found
                    if len(files) > 1:
                        raise Exception(
                            f"too many matching files. please adjust file pattern: {file_pattern}"
                        )
                    file_name = files[0]
                    file_to_upload = f"{download_path}/{file_name}"
                    if s3_prefix.endswith("/"):
                        s3_prefix = s3_prefix[:-1]
                    transfer_config = TransferConfig(
                        multipart_threshold=1024 * 25,
                        max_concurrency=10,
                        multipart_chunksize=1024 * 100,
                        use_threads=True,
                    )
                    self.logger.info(
                        f"uploading file to s3. filename: {file_to_upload} to s3 path: {s3_bucket}/{s3_prefix}/{dest_file_name}"
                    )
                    s3_client.upload_file(
                        Filename=file_to_upload,
                        Bucket=s3_bucket,
                        Key=f"{s3_prefix}/{dest_file_name}",
                        Config=transfer_config,
                    )

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getSourceUrl(self) -> str:
        return self.getOrDefault(self.source_url)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getDownloadToPath(self) -> str:
        return self.getOrDefault(self.download_to_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getExtractZip(self) -> bool:
        return self.getOrDefault(self.extract_zip)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getS3DestinationPath(self) -> str:
        return self.getOrDefault(self.s3_destination_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def GetFileNameMap(self) -> Dict[str, str]:
        return self.getOrDefault(self.file_name_map)



from os import path, makedirs
from pathlib import Path
from shutil import rmtree
from typing import List

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from pyspark.sql.types import StructType

from create_spark_session import clean_spark_session
from spark_pipeline_framework.transformers.framework_json_to_jsonl_converter.v1.framework_json_to_jsonl_converter import (
    FrameworkJsonToJsonlConverter,
)


# noinspection SqlNoDataSourceInspection
def test_can_convert_json_to_jsonl(spark_session: SparkSession) -> None:
    # Arrange
    clean_spark_session(spark_session)

    data_dir: Path = Path(__file__).parent.joinpath("./")
    test_file_path: str = f"{data_dir.joinpath('test_files/test.json')}"

    temp_folder = data_dir.joinpath("temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    # Act
    FrameworkJsonToJsonlConverter(
        file_path=test_file_path, output_folder=temp_folder
    ).transform(df)

    # Assert
    with open(temp_folder.joinpath("test.json"), "r+") as file:
        lines: List[str] = file.readlines()
        assert len(lines) == 2
        assert (
            lines[0]
            == '{"title":"A Philosophy of Software Design","authors":[{"given":["John"],"surname":"Ousterhout"}],"edition":null}\n'
        )
        assert (
            lines[1]
            == '{"title":"Essentials of Programming Languages","authors":[{"given":["Dan","P."],"surname":"Friedman"},{"given":["Mitchell"],"surname":"Wand"}],"edition":3}\n'
        )



import asyncio
import json
from typing import Any, Dict, List, Optional, TYPE_CHECKING

from helixtelemetry.telemetry.structures.telemetry_parent import TelemetryParent
from pyspark.ml.base import Transformer

from pyspark.ml.param import Param

from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable

from pyspark.sql.dataframe import DataFrame
from typing_extensions import final

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.mixins.loop_id_mixin import LoopIdMixin
from spark_pipeline_framework.mixins.telemetry_parent_mixin import TelemetryParentMixin
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.utilities.async_helper.v1.async_helper import AsyncHelper
from spark_pipeline_framework.utilities.class_helpers import ClassHelpers


class FrameworkTransformer(
    Transformer,
    DefaultParamsReadable,  # type: ignore
    DefaultParamsWritable,
    LoopIdMixin,
    TelemetryParentMixin,
):
    def __init__(
        self,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        """
        Base class for our transformers

        """
        super().__init__()

        if TYPE_CHECKING:
            self._input_kwargs: Dict[str, Any] = {}
        self.logger = get_logger(__name__)

        self.name: Param[str] = Param(self, "name", "")
        self._setDefault(name=name)

        self.progress_logger: Param[Optional[ProgressLogger]] = Param(
            self, "progress_logger", ""
        )
        self._setDefault(progress_logger=progress_logger)

        self.loop_id: Optional[str] = None

        self.telemetry_parent: Optional[TelemetryParent] = None

        self.parameters: Optional[Dict[str, Any]] = parameters

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring, PyUnusedLocal
    def setStandardParams(
        self,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ) -> Any:
        kwargs: Dict[str, Any] = self._input_kwargs
        return self._set(**kwargs)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring, PyUnusedLocal,Mypy
    @final
    def setParams(self, **kwargs: Any) -> Any:
        # ignore any parameters
        kwargs = {key: value for key, value in kwargs.items() if self.hasParam(key)}
        return self._set(**kwargs)

    @property
    def transformers(self) -> List[Transformer]:
        return [self]

    # # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    # def setName(self, value: str) -> "FrameworkTransformer":
    #     self._paramMap[self.name] = value
    #     return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getName(self) -> Optional[str]:
        return self.getOrDefault(self.name)

    # # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    # def setProgressLogger(
    #     self, value: Optional[ProgressLogger]
    # ) -> "FrameworkTransformer":
    #     self._paramMap[self.progress_logger] = value
    #     return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getProgressLogger(self) -> Optional[ProgressLogger]:
        return self.getOrDefault(self.progress_logger)

    # # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    # def setParameters(self, value: Optional[Dict[str, Any]]) -> "FrameworkTransformer":
    #     self._paramMap[self.parameters] = value
    #     return self
    #
    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getParameters(self) -> Optional[Dict[str, Any]]:
        return self.parameters

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getSql(self) -> Optional[str]:
        return None

    def _set(self, **kwargs: Any) -> "FrameworkTransformer":
        # filter out any args that don't have parameters
        kwargs = {key: value for key, value in kwargs.items() if self.hasParam(key)}
        # noinspection PyUnresolvedReferences,PyProtectedMember
        super()._set(**kwargs)
        return self

    def __str__(self) -> str:
        return json.dumps(self.as_dict(), default=str)

    def as_dict(self) -> Dict[str, Any]:
        return {
            "name": self.getName(),
            "short_type": self.__class__.__name__,
            "type": ClassHelpers.get_full_name_of_instance(self),
            "params": {
                k.name: (
                    (
                        self.getOrDefault(k)
                        if not hasattr(self.getOrDefault(k), "as_dict")
                        else self.getOrDefault(k).as_dict()
                    )
                    if not callable(self.getOrDefault(k))
                    else ClassHelpers.get_function_as_text(
                        fn=self.getOrDefault(k), strip=f"{k.name}="
                    )
                )
                for k, v in self._paramMap.items()
                if k.name not in ["progress_logger"]
            },
        }

    def update_from_dict(self, dictionary: Dict[str, Any]) -> "FrameworkTransformer":
        if dictionary:
            for key, value in dictionary.items():
                setattr(self, key, value)
        return self

    async def transform_async(self, df: DataFrame) -> DataFrame:
        """
        Transform the dataset asynchronously

        :param df: input dataset
        :return: transformed dataset
        """
        return await self._transform_async(df)

    async def _transform_async(self, df: DataFrame) -> DataFrame:
        """
        Override this method to implement async transformation.  By default, it calls the sync transform method

        :param df: input dataframe
        :return: transformed dataframe
        """

        def transform_wrapper() -> DataFrame:
            return self._transform(df)

        return await asyncio.to_thread(transform_wrapper)

    def _transform(self, df: DataFrame) -> DataFrame:
        """
        Override this method to implement transformation

        :param df: input dataframe
        :return: transformed dataframe
        """
        return AsyncHelper.run(self.transform_async(df))



from pathlib import Path
from typing import Callable, Dict, Any, Union, List

from pyspark.sql import SparkSession
from spark_auto_mapper.automappers.automapper_base import AutoMapperBase

from spark_pipeline_framework.proxy_generator.python_transformer_helpers import (
    get_python_function_from_location,
)


def test_can_find_python_function(spark_session: SparkSession) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath("./")

    # Act
    result_function: Callable[
        [Dict[str, Any]], Union[AutoMapperBase, List[AutoMapperBase]]
    ] = get_python_function_from_location(
        location=str(data_dir.joinpath("library/features/carriers_mapping/v1")),
        import_module_name=".mapping",
        function_name="mapping",
    )

    result: Union[AutoMapperBase, List[AutoMapperBase]] = result_function(
        {"view": "bar"}
    )

    # Assert
    assert isinstance(result, AutoMapperBase)



import traceback
from datetime import datetime
from typing import Any, Dict, Optional
from unittest.mock import AsyncMock

import pytest
from aioresponses import aioresponses
from helix_fhir_client_sdk.exceptions.fhir_sender_exception import FhirSenderException
from helix_fhir_client_sdk.function_types import RefreshTokenResult

from spark_pipeline_framework.transformers.fhir_receiver.v2.fhir_receiver_processor import (
    FhirReceiverProcessor,
)
from spark_pipeline_framework.transformers.fhir_receiver.v2.structures.get_batch_result import (
    GetBatchResult,
)
from spark_pipeline_framework.transformers.fhir_receiver.v2.test.fhir_receiver_processor.get_fhir_receiver_parameters import (
    get_fhir_receiver_parameters,
)


async def test_get_batch_results_paging_async() -> None:
    parameters = get_fhir_receiver_parameters()

    with aioresponses() as m:
        # Mock the FHIR server response
        m.get(
            "http://fhir-server/Patient?_count=5&_getpagesoffset=0",
            payload={
                "resourceType": "Bundle",
                "entry": [{"resource": {"id": "1", "resourceType": "Patient"}}],
            },
        )
        m.get(
            "http://fhir-server/Patient?_count=5&_getpagesoffset=0&id%253Aabove=1",
            payload={
                "resourceType": "Bundle",
                "entry": [{"resource": {"id": "2", "resourceType": "Patient"}}],
            },
        )
        m.get(
            "http://fhir-server/Patient?_count=5&_getpagesoffset=0&id%253Aabove=2",
            status=404,
        )

        loop_number: int = 0
        async for result in FhirReceiverProcessor.get_batch_results_paging_async(
            last_updated_after=None,
            last_updated_before=None,
            limit=10,
            page_size=5,
            parameters=parameters,
            server_url="http://fhir-server",
        ):
            loop_number += 1
            assert isinstance(result, GetBatchResult)
            if loop_number == 1:
                assert len(result.resources) == 1
                assert result.resources[0] == '{"id": "1", "resourceType": "Patient"}'
            elif loop_number == 2:
                assert len(result.resources) == 1
                assert result.resources[0] == '{"id": "2", "resourceType": "Patient"}'
            else:
                assert len(result.resources) == 0


async def test_get_batch_results_paging_empty_bundle_async() -> None:
    parameters = get_fhir_receiver_parameters()

    with aioresponses() as m:
        # Mock the FHIR server response
        m.get(
            "http://fhir-server/Patient?_count=5&_getpagesoffset=0",
            payload={
                "resourceType": "Bundle",
                "entry": [{"resource": {"id": "1", "resourceType": "Patient"}}],
            },
        )
        m.get(
            "http://fhir-server/Patient?_count=5&_getpagesoffset=0&id%253Aabove=1",
            payload={
                "resourceType": "Bundle",
                "entry": [{"resource": {"id": "2", "resourceType": "Patient"}}],
            },
        )
        m.get(
            "http://fhir-server/Patient?_count=5&_getpagesoffset=0&id%253Aabove=2",
            payload={
                "resourceType": "Bundle",
                "type": "searchset",
                "timestamp": "2022-08-08T07:50:38.3838Z",
                "total": 0,
                "entry": [],
            },
        )

        loop_number: int = 0
        async for result in FhirReceiverProcessor.get_batch_results_paging_async(
            last_updated_after=None,
            last_updated_before=None,
            limit=10,
            page_size=5,
            parameters=parameters,
            server_url="http://fhir-server",
        ):
            loop_number += 1
            assert isinstance(result, GetBatchResult)
            if loop_number == 1:
                assert len(result.resources) == 1
                assert result.resources[0] == '{"id": "1", "resourceType": "Patient"}'
            elif loop_number == 2:
                assert len(result.resources) == 1
                assert result.resources[0] == '{"id": "2", "resourceType": "Patient"}'
            else:
                assert len(result.resources) == 0


async def test_get_batch_result_streaming_async_no_resources() -> None:
    parameters = get_fhir_receiver_parameters()
    with aioresponses() as m:
        m.get(
            "http://fhir-server/Patient",
            payload={
                "resourceType": "Bundle",
                "type": "searchset",
                "total": 0,
                "entry": [],
            },
        )

        async_gen = FhirReceiverProcessor.get_batch_result_streaming_async(
            last_updated_after=None,
            last_updated_before=None,
            parameters=parameters,
            server_url="http://fhir-server/",
        )

        results = [result async for result in async_gen]
        assert len(results) == 1
        assert results[0]["resources"] == []
        assert results[0]["errors"] == []


async def test_get_batch_result_streaming_async_with_resources() -> None:
    parameters = get_fhir_receiver_parameters()
    with aioresponses() as m:
        m.get(
            "http://fhir-server/Patient",
            payload={
                "resourceType": "Bundle",
                "type": "searchset",
                "total": 1,
                "entry": [{"resource": {"resourceType": "Patient", "id": "1"}}],
            },
        )
        m.get("http://fhir-server/Patient?id:above=1", payload=[])

        async_gen = FhirReceiverProcessor.get_batch_result_streaming_async(
            last_updated_after=None,
            last_updated_before=None,
            parameters=parameters,
            server_url="http://fhir-server/",
        )

        results = [result async for result in async_gen]
        assert len(results) == 2
        assert results[0]["resources"][0] == '{"resourceType": "Patient", "id": "1"}'
        assert results[0]["errors"] == []


async def test_get_batch_result_streaming_async_with_error() -> None:
    parameters = get_fhir_receiver_parameters()
    with aioresponses() as m:
        m.get(
            "http://fhir-server/Patient",
            status=500,
            payload={
                "resourceType": "OperationOutcome",
                "issue": [
                    {
                        "severity": "error",
                        "code": "exception",
                        "diagnostics": "Internal Server Error",
                    }
                ],
            },
        )
        m.get(
            "http://fhir-server/Patient",
            status=500,
            payload={
                "resourceType": "OperationOutcome",
                "issue": [
                    {
                        "severity": "error",
                        "code": "exception",
                        "diagnostics": "Internal Server Error",
                    }
                ],
            },
        )

        parameters.retry_count = 1

        with pytest.raises(FhirSenderException):
            async_gen = FhirReceiverProcessor.get_batch_result_streaming_async(
                last_updated_after=None,
                last_updated_before=None,
                parameters=parameters,
                server_url="http://fhir-server/",
            )
            assert [result async for result in async_gen] == []


async def test_get_batch_result_streaming_async_with_error_with_retry_success() -> None:
    parameters = get_fhir_receiver_parameters()
    with aioresponses() as m:
        m.get(
            "http://fhir-server/Patient",
            status=500,
            payload={
                "resourceType": "OperationOutcome",
                "issue": [
                    {
                        "severity": "error",
                        "code": "exception",
                        "diagnostics": "Internal Server Error",
                    }
                ],
            },
        )
        m.get(
            "http://fhir-server/Patient",
            payload={"resourceType": "Patient", "id": "1"},
        )
        m.get("http://fhir-server/Patient?id:above=1", payload=[])

        result: Dict[str, Any]
        async_gen = FhirReceiverProcessor.get_batch_result_streaming_async(
            last_updated_after=None,
            last_updated_before=None,
            parameters=parameters,
            server_url="http://fhir-server/",
        )

        results = [result async for result in async_gen]
        assert len(results) == 2
        assert results[0]["resources"][0] == '{"resourceType": "Patient", "id": "1"}'
        assert results[0]["errors"] == []


async def test_get_batch_result_streaming_async_with_auth_error_with_re_auth() -> None:
    print()

    parameters = get_fhir_receiver_parameters()
    with aioresponses() as m:
        m.get("http://fhir-server/Patient", status=401)
        m.get(
            "http://fhir-server/Patient",
            payload={"resourceType": "Patient", "id": "1"},
        )
        m.get("http://fhir-server/Patient?id:above=1", payload=[])

        # noinspection PyUnusedLocal
        async def my_refresh_token_function(
            url: Optional[str],
            status_code: Optional[int],
            current_token: Optional[str],
            expiry_date: Optional[datetime],
            retry_count: Optional[int],
        ) -> RefreshTokenResult:
            print("Call stack:")
            traceback.print_stack()
            return RefreshTokenResult(
                access_token="new_token", expiry_date=None, abort_request=False
            )

        mock_refresh_token_function = AsyncMock(side_effect=my_refresh_token_function)

        parameters.refresh_token_function = mock_refresh_token_function
        parameters.auth_access_token = "old_token"

        result: Dict[str, Any]
        async_gen = FhirReceiverProcessor.get_batch_result_streaming_async(
            last_updated_after=None,
            last_updated_before=None,
            parameters=parameters,
            server_url="http://fhir-server/",
        )

        results = [result async for result in async_gen]
        assert len(results) == 2
        # Asserting the first result only, since the second one is empty.
        assert results[0]["resources"][0] == '{"resourceType": "Patient", "id": "1"}'
        assert results[0]["errors"] == []


async def test_get_batch_result_streaming_async_not_found() -> None:
    parameters = get_fhir_receiver_parameters()
    with aioresponses() as m:
        m.get("http://fhir-server/Patient", status=404)

        result: Dict[str, Any]
        async for result in FhirReceiverProcessor.get_batch_result_streaming_async(
            last_updated_after=None,
            last_updated_before=None,
            parameters=parameters,
            server_url="http://fhir-server/",
        ):
            assert isinstance(result, dict)
            assert result["resources"] == []



from os import path, makedirs
from pathlib import Path
from shutil import rmtree
from typing import Any, Tuple, Dict, Iterator, Optional

from mockserver_client.mock_requests_loader import load_mock_source_api_json_responses
from mockserver_client.mockserver_client import MockServerFriendlyClient
from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import StructType, StructField, StringType
from requests import Response

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.http_data_receiver.v4.http_data_receiver import (
    HttpDataReceiver,
)
from spark_pipeline_framework.utilities.api_helper.http_request import HelixHttpRequest
from spark_pipeline_framework.utilities.oauth2_helpers.v2.oauth2_client_credentials_flow import (
    OAuth2Credentails,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)


def test_http_data_receiver(spark_session: SparkSession) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath("./")
    test_name = "test_http_data_receiver"

    # remove temp directory
    temp_folder = data_dir.joinpath("./temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    df: DataFrame = create_empty_dataframe(spark_session=spark_session)

    server_url = f"http://mock-server:1080/{test_name}"
    mock_client = MockServerFriendlyClient("http://mock-server:1080")
    mock_client.clear(test_name)
    mock_client.reset()

    load_mock_source_api_json_responses(
        folder=data_dir.joinpath("oauth_2_service"),
        mock_client=mock_client,
        url_prefix=f"{test_name}/token",
    )

    load_mock_source_api_json_responses(
        folder=data_dir.joinpath("token_service"),
        mock_client=mock_client,
        url_prefix=f"{test_name}/eligibility",
    )

    def http_request_generator(
        _: DataFrame, _1: Optional[ProgressLogger]
    ) -> Iterator[Tuple[HelixHttpRequest, int]]:
        for i in range(2):
            yield HelixHttpRequest(
                url=f"{server_url}/eligibility",
                headers={
                    "Content-Type": "application/json",
                },
            ), i

    def response_processor(
        response: Response, state: Any
    ) -> Tuple[Dict[str, Any], bool]:
        assert isinstance(state, int)
        # Considering state 2 as error state
        return response.json(), state == 1

    # Act
    with ProgressLogger() as progress_logger:
        HttpDataReceiver(
            name="Testing HTTP Data Receiver",
            success_view="success_view",
            error_view="error_view",
            http_request_generator=http_request_generator,
            response_processor=response_processor,
            success_schema=StructType(
                [
                    StructField("token_type", StringType()),
                ]
            ),
            credentials=OAuth2Credentails(
                client_id="client_id", client_secret="client_secret"
            ),
            auth_url=f"{server_url}/token",
            batch_size=15,
            run_sync=True,
        ).transform(df)

    # Assert
    success_df: DataFrame = spark_session.table("success_view")
    success_df.printSchema()
    success_df.show(truncate=False)

    assert success_df.collect()[0].asDict(recursive=True) == {
        "headers": {"Authorization": "Bearer access_token_1"},
        "url": "http://mock-server:1080/test_http_data_receiver/eligibility",
        "status": 200,
        "state": "0",
        "data": {"token_type": "bearer"},
    }

    error_df: DataFrame = spark_session.table("error_view")
    error_df.printSchema()
    error_df.show(truncate=False)

    assert error_df.collect()[0].asDict(recursive=True) == {
        "headers": {"Authorization": "Bearer access_token_1"},
        "url": "http://mock-server:1080/test_http_data_receiver/eligibility",
        "status": 200,
        "state": "1",
        "data": {"token_type": "bearer"},
    }



import json
from os import environ, path, makedirs
from pathlib import Path
from shutil import rmtree
from typing import Optional

import pytest
from helix_fhir_client_sdk.responses.fhir_merge_response import FhirMergeResponse
from helix_fhir_client_sdk.structures.get_access_token_result import (
    GetAccessTokenResult,
)
from helix_fhir_client_sdk.utilities.fhir_helper import FhirHelper
from helix_fhir_client_sdk.utilities.fhir_server_helpers import FhirServerHelpers
from pyspark.sql import DataFrame, SparkSession

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.fhir_receiver.v2.fhir_receiver import (
    FhirReceiver,
)
from spark_pipeline_framework.utilities.fhir_server_test_context.v1.fhir_server_test_context import (
    FhirServerTestContext,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)


@pytest.mark.parametrize("run_synchronously", [True, False])
@pytest.mark.parametrize("use_data_streaming", [True, False])
async def test_async_real_fhir_server_get_patients_large_with_limit(
    spark_session: SparkSession, run_synchronously: bool, use_data_streaming: bool
) -> None:
    print()
    print(
        f"run_synchronously: {run_synchronously}, use_data_streaming: {use_data_streaming}"
    )
    data_dir: Path = Path(__file__).parent.joinpath("./")

    temp_folder = data_dir.joinpath("../temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    patient_json_path: Path = temp_folder.joinpath("patient.json")

    resource_type = "Patient"
    async with FhirServerTestContext(
        resource_type="Patient"
    ) as fhir_server_test_context:

        environ["LOGLEVEL"] = "DEBUG"

        fhir_server_url: str = fhir_server_test_context.fhir_server_url
        auth_client_id = fhir_server_test_context.auth_client_id
        auth_client_secret = fhir_server_test_context.auth_client_secret
        auth_well_known_url = fhir_server_test_context.auth_well_known_url

        fhir_client = await fhir_server_test_context.create_fhir_client_async()

        await FhirServerHelpers.clean_fhir_server_async(
            resource_type=resource_type, owner_tag="medstar"
        )

        fhir_client = fhir_client.url(fhir_server_url).resource(resource_type)

        count = 10

        print(f"Deleting {count} patients")
        patient_ids = [f"example-{i}" for i in range(count)]
        await FhirHelper.delete_resources_by_ids_async(
            fhir_client=fhir_client, resource_type=resource_type, id_list=patient_ids
        )
        print(f"Deleted {count} patients")

        fhir_client = await fhir_server_test_context.create_fhir_client_async()
        fhir_client = fhir_client.url(fhir_server_url).resource(resource_type)

        resource = await FhirHelper.create_test_patients(count)
        print(f"Merging {count} patients")
        merge_response: Optional[FhirMergeResponse] = (
            await FhirMergeResponse.from_async_generator(
                fhir_client.merge_async(json_data_list=[json.dumps(resource)])
            )
        )
        assert merge_response is not None
        print(f"Merged {count} patients")
        print(merge_response.responses)
        assert merge_response.status == 200, merge_response.responses
        assert len(merge_response.responses) == count, merge_response.responses
        assert merge_response.responses[0]["created"] is True, merge_response.responses

        logger = get_logger(__name__)

        access_token_result: GetAccessTokenResult = (
            await fhir_server_test_context.get_access_token_async()
        )
        access_token: Optional[str] = access_token_result.access_token
        print(f"Found access token in test: {access_token}")
        assert access_token is not None
        assert isinstance(access_token, str)
        # act
        df: DataFrame = create_empty_dataframe(spark_session=spark_session)

        parameters = {"flow_name": "Test Pipeline V2", "team_name": "Data Operations"}

        with ProgressLogger() as progress_logger:
            await FhirReceiver(
                server_url=fhir_server_url,
                resource=resource_type,
                file_path=patient_json_path,
                progress_logger=progress_logger,
                parameters=parameters,
                run_synchronously=run_synchronously,
                auth_well_known_url=auth_well_known_url,
                auth_client_id=auth_client_id,
                auth_client_secret=auth_client_secret,
                use_data_streaming=use_data_streaming,
                limit=5,
                use_uuid_for_id_above=True,
            ).transform_async(df)

        # Assert
        json_df: DataFrame = df.sparkSession.read.json(str(patient_json_path))
        json_df.show(truncate=False, n=50)
        json_df.printSchema()

        assert json_df.count() == 5



import json
from typing import Any, Dict, Iterable, List, Optional, Union, Callable, Tuple, cast
from functools import partial

from requests import status_codes, Response
from pyspark.sql.types import Row
from pyspark import SparkFiles

from spark_pipeline_framework.register import register
from spark_pipeline_framework.utilities.api_helper.http_request import (
    HelixHttpRequest,
    RequestType,
)
from spark_pipeline_framework.utilities.oauth2_helpers.v1.oauth2_client_credentials_flow import (
    OAuth2ClientCredentialsFlow,
)


class HttpDataSenderProcessor:
    @staticmethod
    def create_access_token(client_id: str, auth_url: str, client_secret: str) -> str:
        """
        Factory function to create access token

        :param client_id: OAuth client id
        :param auth_url: OAuth token URL
        :param client_secret: OAuth client secret
        """
        oauth2_client_credentials_flow: OAuth2ClientCredentialsFlow = (
            OAuth2ClientCredentialsFlow(
                auth_url=auth_url,
                client_id=client_id,
                client_secret=client_secret,
                progress_logger=None,
            )
        )

        access_token: Optional[str] = oauth2_client_credentials_flow.get_token()
        assert access_token
        print(f"Received token from {auth_url}: {access_token}")

        return access_token

    @staticmethod
    def create_request(
        url: Optional[str],
        headers: Dict[str, Any],
        raise_error: bool,
        post_as_json_formatted_string: Optional[bool],
        json_data: Dict[str, Any],
        payload_generator: Optional[Callable[[Dict[str, Any]], Dict[str, Any]]],
        url_generator: Optional[Callable[[Dict[str, Any]], str]],
        cert: Optional[Union[str, Tuple[str, str]]],
        verify: Optional[Union[bool, str]],
    ) -> Response:
        """
        Function to initiate the request and create row

        :param url: target url
        :param headers: headers for the request
        :param raise_error: flag to raise error if request fails
        :param post_as_json_formatted_string: flag to convert the json to json string
        :param json_data: payload for the API
        :param payload_generator: function to create the payload
        :param url_generator: function to create the url
        :param cert: certificate or ca bundle file path
        :param verify: controls whether the SSL certificate of the server should be verified when making HTTPS requests.
        """

        url = cast(str, url_generator(json_data) if url_generator else url)
        request: HelixHttpRequest = HelixHttpRequest(
            request_type=RequestType.POST,
            url=url,
            headers=headers,
            payload=payload_generator(json_data) if payload_generator else json_data,
            post_as_json_formatted_string=post_as_json_formatted_string,
            cert=cert,
            verify=verify,
            raise_error=raise_error,
        )
        return request.get_response()

    @staticmethod
    # function that is called for each partition
    # noinspection PyUnusedLocal
    def send_partition_to_server(
        *,
        partition_index: int,
        rows: Iterable[Row],
        raise_error: bool,
        url: Optional[str],
        content_type: str,
        headers: Dict[str, Any],
        post_as_json_formatted_string: Optional[bool],
        client_id: Optional[str],
        auth_url: Optional[str],
        client_secret: Optional[str],
        payload_generator: Optional[Callable[[Dict[str, Any]], Dict[str, Any]]],
        url_generator: Optional[Callable[[Dict[str, Any]], str]],
        response_processor: Optional[Callable[[Dict[str, Any], Response], Any]],
        cert: Optional[Union[str, Tuple[str, str]]],
        verify: Optional[Union[bool, str]],
    ) -> Iterable[Row]:
        """
        This function processes a partition

        This has to be a static function to avoid creating a closure around a class
        https://spark.apache.org/docs/latest/rdd-programming-guide.html#passing-functions-to-spark
        """
        register()
        oauth_enabled = bool(client_id and auth_url and client_secret)

        json_data_list: List[Dict[str, Any]] = [r.asDict(recursive=True) for r in rows]
        print(
            f"----- Sending batch {partition_index} containing {len(json_data_list)} rows -----"
        )

        # logger = get_logger(__name__)
        if len(json_data_list) == 0:
            yield Row(
                url="",
                status=0,
                is_error=False,
                error_data="",
                success_data="",
                headers={},
                request_type="",
            )

        # Assumes certs are distributed to the executors beforehand via SparkContext.addFile
        cert_files: Optional[Union[str, Tuple[str, str]]] = None
        if isinstance(cert, tuple):
            cert_files = SparkFiles.get(cert[0]), SparkFiles.get(cert[1])
        elif cert:
            cert_files = SparkFiles.get(cert)

        headers["Content-Type"] = content_type
        if oauth_enabled:
            assert client_id
            assert auth_url
            assert client_secret
            access_token = HttpDataSenderProcessor.create_access_token(
                client_id, auth_url, client_secret
            )
            headers.update({"Authorization": f"Bearer {access_token}"})

        assert url or url_generator
        json_data: Dict[str, Any]
        for json_data in json_data_list:
            create_request = partial(
                HttpDataSenderProcessor.create_request,
                url=url,
                raise_error=False,
                post_as_json_formatted_string=post_as_json_formatted_string,
                json_data=json_data,
                payload_generator=payload_generator,
                url_generator=url_generator,
                cert=cert_files,
                verify=verify,
            )
            response: Response = create_request(headers=headers)

            if (
                oauth_enabled
                and response.status_code == status_codes.codes.unauthorized
            ):
                assert client_id
                assert auth_url
                assert client_secret
                access_token = HttpDataSenderProcessor.create_access_token(
                    client_id, auth_url, client_secret
                )
                headers.update({"Authorization": f"Bearer {access_token}"})
                response = create_request(headers=headers, raise_error=raise_error)
            elif raise_error:
                response.raise_for_status()

            if response_processor:
                result, is_error = response_processor(json_data, response)
            else:
                result, is_error = response.json(), response.status_code >= 400

            yield Row(
                url=response.url,
                status=response.status_code,
                is_error=is_error,
                error_data=json.dumps(result if is_error else None),
                success_data=json.dumps(None if is_error else result),
                headers=headers,
                request_type=str(RequestType.POST),
            )



from typing import Optional, List, cast

from pyspark.sql.types import (
    Row,
)

from spark_pipeline_framework.utilities.fhir_helpers.fhir_merge_response_schema import (
    FhirMergeResponseSchema,
)


class FhirMergeResponseReader:
    def __init__(self, row: Row) -> None:
        """
        Reads a Row object and allows property access to it


        :param row: A Spark Row
        """
        self.row: Row = row

    @property
    def request_id(self) -> Optional[str]:
        return self._safe_get(FhirMergeResponseSchema.request_id)

    @property
    def url(self) -> str:
        return cast(str, self.row[FhirMergeResponseSchema.request_id])

    @property
    def responses(self) -> List[str]:
        return cast(List[str], self.row[FhirMergeResponseSchema.responses])

    @property
    def error(self) -> Optional[str]:
        return self._safe_get(FhirMergeResponseSchema.error)

    @property
    def access_token(self) -> Optional[str]:
        return self._safe_get(FhirMergeResponseSchema.access_token)

    @property
    def status(self) -> int:
        return cast(int, self.row[FhirMergeResponseSchema.status])

    @property
    def data(self) -> str:
        return cast(str, self.row[FhirMergeResponseSchema.data])

    def _safe_get(self, property_name: str) -> Optional[str]:
        return self.row[property_name] if property_name in self.row else None



from typing import List, Tuple

import pytest
from aioresponses import aioresponses
from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import StructType, StructField, StringType

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.elasticsearch_sender.v2.elasticsearch_sender import (
    ElasticSearchSender,
)


async def test_transform_async(spark_session: SparkSession) -> None:
    with ProgressLogger() as progress_logger:
        data = [
            {"id": 1, "name": "Alice"},
            {"id": 2, "name": "Bob"},
        ]
        schema = StructType(
            [
                StructField("id", StringType()),
                StructField("name", StringType()),
            ]
        )
        df = spark_session.createDataFrame(data, schema=schema)  # type: ignore[type-var]
        df.createOrReplaceTempView("test_view")
        sender = ElasticSearchSender(
            index="test_index",
            file_path=None,
            view="test_view",
            progress_logger=progress_logger,
            run_synchronously=True,
        )

        with aioresponses() as m:
            return_payload = {
                "items": [
                    {
                        "index": {
                            "_index": "test",
                            "_id": "1",
                            "_version": 1,
                            "result": "created",
                            "_shards": {"total": 2, "successful": 2, "failed": 0},
                            "status": 201,
                            "_seq_no": 0,
                            "_primary_term": 1,
                        }
                    },
                    {
                        "index": {
                            "_index": "test",
                            "_id": "2",
                            "_version": 1,
                            "result": "created",
                            "_shards": {"total": 2, "successful": 2, "failed": 0},
                            "status": 201,
                            "_seq_no": 0,
                            "_primary_term": 1,
                        }
                    },
                ],
                "errors": False,
            }

            m.post(
                "https://elasticsearch:9200/test_index/_bulk",
                status=200,
                payload=return_payload,
            )

            result_df = await sender._transform_async(df)
            result_df.show(truncate=False)
            assert result_df.count() == 1
            assert result_df.collect()[0]["success"] == 2
            assert result_df.collect()[0]["failed"] == 0


async def test_transform_async_with_errors(spark_session: SparkSession) -> None:
    with ProgressLogger() as progress_logger:
        data = [
            {"id": 1, "name": "Alice"},
            {"id": 2, "name": "Bob"},
        ]
        schema = StructType(
            [
                StructField("id", StringType()),
                StructField("name", StringType()),
            ]
        )
        df = spark_session.createDataFrame(data, schema=schema)  # type: ignore[type-var]
        df.createOrReplaceTempView("test_view")
        sender = ElasticSearchSender(
            index="test_index",
            file_path=None,
            view="test_view",
            progress_logger=progress_logger,
            run_synchronously=True,
        )

        with aioresponses() as m:
            m.post(
                "https://elasticsearch:9200/test_index/_bulk",
                status=500,
                payload={"error": "Internal Server Error"},
            )

            with pytest.raises(Exception):
                result_df = await sender._transform_async(df)
                result_df.show(truncate=False)
                assert result_df.count() == 1
                assert result_df.collect()[0]["success"] == 0
                assert result_df.collect()[0]["failed"] == 2


@pytest.mark.parametrize("run_synchronously", [True, False])
def test_elasticsearch_sender_real_elasticsearch(
    spark_session: SparkSession, run_synchronously: bool
) -> None:
    # Given
    data: List[Tuple[str, str]] = [
        ("1", "{'id': '1', 'name': 'name1'}"),
        ("2", "{'id': '2', 'name': 'name2'}"),
    ]
    schema: StructType = StructType(
        [StructField("id", StringType()), StructField("json", StringType())]
    )
    df: DataFrame = spark_session.createDataFrame(data, schema)

    # use just one partition to match the synchronous test
    df = df.coalesce(1)
    view: str = "test_view"
    df.createOrReplaceTempView(view)

    elasticsearch_sender: ElasticSearchSender = ElasticSearchSender(
        name="elasticsearch_sender",
        view=view,
        index="test_index",
        operation="index",
        parameters={"doc_id_prefix": "test"},
        progress_logger=ProgressLogger(),
        run_synchronously=run_synchronously,
    )

    # When
    result_df: DataFrame = elasticsearch_sender.transform(df)

    # Then
    result_df.show(truncate=False)

    assert result_df.count() == 1
    assert result_df.columns == [
        "url",
        "success",
        "failed",
        "payload",
        "partition_index",
        "error",
    ]
    assert result_df.collect()[0]["success"] == 2
    assert result_df.collect()[0]["failed"] == 0
    assert result_df.collect()[0]["url"] == "https://elasticsearch:9200/test_index"
